/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <unistd.h>
#include <sys/types.h>

#include <QDebug>
#include <QStringList>

#include <QApplication>
#include <QTranslator>
#include <QWidget>

#include "WidgetManager.h"
#include "widgetinstallerview.h"

#include "wbenchmark.h"

void printExitMessage(int result) {
    qDebug() << "*************************************************";
    qDebug() << "WidgetInstaller exited with " << result;
    qDebug() << "*************************************************";
}

void printHelp() {
    QTextStream out(stdout);

    out << "Usage: widgetinstaller WIDGET [OPTIONS]\n";
    out << "   [--silent]  Run widgetinstaller in silent mode\n";
    out << "   [--no-gui]  Run widgetinstaller without GUI\n";
    out << "   [--no-install]  Convert widget to Debian package but don't install it\n";
    out << "   [--path] Path to write the converted Debian package\n";
    out << "   [--ovi-productid] name of the debian package or productId\n";
}

int main(int argc, char *argv[])
{
    WTS("widgetinstaller started at");
    WBM("widgetinstaller");

    QStringList argsList;
    for (int i = 0; i < argc; i++) {
        argsList << QString(argv[i]);
    }

    if (argsList.contains("--help") || argsList.contains("-h")){
        printHelp();
        return 0;
    }

    QString logMsg;
    logMsg = "widgetinstaller log, created ";
    logMsg += QDateTime::currentDateTime().toString("ddd MMM hh:mm:ss yyyy");
    logMsg += "\nCommand: " + argsList.join(" ")+"\n";

    bool packageOnly = false;
    if (argsList.contains("--path")) {
        int fileIndex = argsList.indexOf("--path")+1;
        if (argsList.size() > fileIndex) {
            QDir debianPath(argsList.at(fileIndex));
            if(!debianPath.exists() || !QFileInfo(debianPath.absolutePath()).isWritable()) {
                logMsg = "Path to write Debian package doesn't exist or the directory is not writable";
                qCritical() << logMsg;
                return 4;
            } else {
                packageOnly = true;
            }
        }
    }

    if (geteuid() == 0 && !(argsList.contains("--allow-preinstall") || argsList.contains("--allow-root"))) {
        logMsg = "Running as root! You don't want to do this!\n";
        logMsg += "Use --allow-root if you really know what you are doing.";
        qCritical() << logMsg;
        return 2;
    }

    if (argc < 2) {
        printHelp();
        return 1;
    }

    QFile widgetpath(argv[1]);
    if (!widgetpath.exists()) {
        logMsg = ".Wgt package doesn't exist";
        qCritical() <<  logMsg;
        return 3;
    }

    bool silentflag = argsList.contains("--silent");
    bool noguiflag  = argsList.contains("--no-gui");

    WRT::InstallationAttributes attributes = WRT::NoAttibutes;
    if (argsList.contains("--no-install")) {
        attributes |= WRT::NoInstall;
        packageOnly = true;
    }

    QString widgetId;

    if (noguiflag) {
        QCoreApplication app(argc, argv);
        WidgetManager mgr(NULL);
        mgr.setInstallationAttributes(attributes);
        int err = mgr.install(argsList[1], widgetId, true);
        printExitMessage(err);
        return err;
    }

    QApplication app(argc, argv);
    QTranslator translator;
    translator.load(":/widgetinstaller.qm");
    app.installTranslator(&translator);
    QWidget window;

    WidgetManager mgr(NULL);
    mgr.setInstallationAttributes(attributes);

    WidgetInstallerView view(&mgr);
    view.setSilentMode(silentflag);

#ifndef Q_OS_MAEMO5
    // show main window only if silent flag is not set
    if (!silentflag)
        window.show();
#endif

    qRegisterMetaType<WidgetInstallError>();
    QObject::connect(&mgr, SIGNAL(installationError(WidgetInstallError)),
                     &view, SLOT(slotInstallationState(WidgetInstallError)));
    QObject::connect(&mgr, SIGNAL(aboutToReplaceExistingWidget(QString)),
                     &view, SLOT(slotAboutToReplaceExistingWidget()));
    QObject::connect(&mgr, SIGNAL(aboutToInstallWidgetWithFeatures(QList<QString>)),
                     &view, SLOT(slotAboutToInstallWidgetWithFeatures(QList<QString>)));
    QObject::connect(&mgr, SIGNAL(installationSucceed()),
                     &view, SLOT(slotInstallationSucceed()));
    QObject::connect(&mgr, SIGNAL(aboutToInstallUntrustedWidget(WidgetInstallError)),
                     &view, SLOT(slotAboutToInstallUntrustedWidget(WidgetInstallError)));
    QObject::connect(&view, SIGNAL(installationCancelled()),
                     &mgr, SLOT(cancelInstallation()),
                     Qt::DirectConnection);

    int retval = mgr.install(argsList[1], widgetId, true);

    if (silentflag)
        printExitMessage(retval);
    else
        qWarning()<<"Installation failed to errorcode:"<<retval;

    return retval;
}
