/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "wrtgraphicswebview.h"
#include "runtimeglobal.h"
#include "wrtpage.h"
#include <QApplication>
#include <QInputContext>

namespace WRT {

WrtGraphicsWebView::WrtGraphicsWebView(QGraphicsItem* parent)
    : QGraphicsWebView(parent)
    , m_contextMenuEnabled(false)
{
    connect(this, SIGNAL(loadStarted()), this, SLOT(handleLoadStarted()));
}

WrtGraphicsWebView::~WrtGraphicsWebView()
{
}

void WrtGraphicsWebView::enableContextMenu(bool enable)
{
    m_contextMenuEnabled = enable;
}

void WrtGraphicsWebView::contextMenuEvent(QGraphicsSceneContextMenuEvent* ev)
{
    if (m_contextMenuEnabled)
    {
        // To implement a custom context menu don't call the base class' implemenation
        // For more information please see WrtGraphicsWebView's documentation
        QGraphicsWebView::contextMenuEvent(ev);
    }
}


bool WrtGraphicsWebView::event(QEvent* event)
{
    switch (event->type()) {
        case QEvent::InputMethod:   { // The printable key events are passed through this case
            inputMethodEvent(static_cast<QInputMethodEvent*>(event));
            return true;
        }
        case QEvent::KeyPress:      { // We are filtering here to pass the non-printable keyPress events
            QKeyEvent *ev= static_cast<QKeyEvent*>(event);
            keyPressEvent(ev);
            return true;
        }
        case QEvent::KeyRelease:    { // We are filtering here to pass the non-printable keyRelease events
            QKeyEvent *ev= static_cast<QKeyEvent*>(event);
            keyReleaseEvent(ev);
            return true;
        }
        default: {
            qt_noop();
            break;
        }
    }// END switch (event->type())

    return QGraphicsWebView::event(event);
}

void WrtGraphicsWebView::copy()
{
    // QGraphicsWebView returns always a valid pointer to page.
    page()->action(QWebPage::Copy)->trigger();
}

void WrtGraphicsWebView::paste()
{
    // QGraphicsWebView returns always a valid pointer to page.
    page()->action(QWebPage::Paste)->trigger();
}

void WrtGraphicsWebView::cut()
{
    // QGraphicsWebView returns always a valid pointer to page.
    page()->action(QWebPage::Cut)->trigger();
}

void WrtGraphicsWebView::handleSelectionChanged()
{
    emit copyAvailable(!page()->selectedText().isEmpty());
}

QVariant WrtGraphicsWebView::itemChange(GraphicsItemChange change, const QVariant& value)
{
    if (change == QGraphicsItem::ItemFlagsHaveChanged){
        bool ok;
        uint flags = value.toUInt(&ok);
        if (ok & !(flags & QGraphicsItem::ItemAcceptsInputMethod)) {
            QInputContext *ic = qApp->inputContext();
            if (ic) {
                QEvent *closeEvent = new QEvent(QEvent::CloseSoftwareInputPanel);
                ic->filterEvent(closeEvent);
                delete closeEvent;
            }
        }
    }
    return QGraphicsWebView::itemChange(change,value);
}

void WrtGraphicsWebView::handleLoadStarted()
{
    //Navigation has started so dismiss the input method.
    setFlag(QGraphicsItem::ItemAcceptsInputMethod, false);
    disconnect(this, SLOT(handleSelectionChanged()));
    connect(page(), SIGNAL(selectionChanged()), this, SLOT(handleSelectionChanged()));
}

}
