/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "wrtpage.h"
#include "qwebhistory.h"
#include "wrtcontroller.h"
#include "wrtpage_p.h"
#include "wrtsettings.h"
#include "../widgetmanager/src/widgetmanagerconstants.h"

#include <QAbstractButton>
#include <QInputDialog>

#include <QMessageBox>
#include <QTextStream>
#include <QFile>
#include <QDebug>
#include <QEventLoop>
#include <QGraphicsWidget>

#include <QDesktopServices>
#include <QDir>

void QWEBKIT_EXPORT qt_webpage_setGroupName(QWebPage* page, const QString& groupName);
static const qint64 OFFLINE_WEBAPP_DEFAULT_QUOTA  = 5000000;

namespace WRT {

WrtPage::WrtPage(QWidget *parent)
    : QWebPage(parent)
    , d(new WrtPagePrivate())
{
    settings()->setAttribute(QWebSettings::PluginsEnabled, true);
    // Initialization below can only be done once per Webkit instance (and not once per page)
    // otherwise it hits an ASSERT in WebKit in debug mode
    static bool initialized = false;
    // Debug mode can be dealt with in debug - we need these settings while opening a database
    // while another widget is running in the backgound. That is why: if (!initialized) is commented.
    //   if (!initialized) {
        /* calculating font size: taking the dafult (e.g. 16) QWebSettings::DefaultFontSize
         * and using the four input size (smallest;small;normal;large;largest -> 0-4) to adjust it
        */
        int calculatedFontSize =((settings()->fontSize(QWebSettings::DefaultFontSize)) * (WRT::WrtSettings::createWrtSettings()->valueAsInt("DefaultTextSize"))/2);
        // in case we are enum 0, i.e. "smallest", the font size result will be 0: need to adjust a bit
        if (!calculatedFontSize)
            calculatedFontSize = settings()->fontSize(QWebSettings::DefaultFontSize)/4;
        // make sure it is not smaller then the minimum we want to support (default parameters)
        if (calculatedFontSize < settings()->fontSize(QWebSettings::MinimumFontSize))
            calculatedFontSize = settings()->fontSize(QWebSettings::MinimumFontSize);
        // set both types to this value
        settings()->setFontSize (QWebSettings::DefaultFontSize, calculatedFontSize);
        settings()->setFontSize (QWebSettings::DefaultFixedFontSize, calculatedFontSize);
        initialized = true;
    //   }
#ifdef QTWRT_ENABLE_NOTIFICATIONS
    connect(this, SIGNAL(requestPermissionFromUser(QWebFrame*, QWebPage::PermissionDomain)), this, SLOT(requestPermission(QWebFrame*, QWebPage::PermissionDomain)));
    connect(this, SIGNAL(checkPermissionFromUser(QWebFrame*, QWebPage::PermissionDomain, QWebPage::PermissionPolicy&)), this, SLOT(checkPermission(QWebFrame*, QWebPage::PermissionDomain, QWebPage::PermissionPolicy&)));
    connect(this, SIGNAL(cancelRequestsForPermission(QWebFrame*, QWebPage::PermissionDomain)), this, SLOT(cancelRequestsForPermission(QWebFrame*, QWebPage::PermissionDomain)));
#endif
}

WrtPage::~WrtPage()
{
    cancelDialog();
    delete d;
}

void WrtPage::setPageGroupName(const QString& groupName)
{
    qt_webpage_setGroupName(this, groupName);
}

QWebPage* WrtPage::createWindow(QWebPage::WebWindowType webWindowType)
{

    Q_UNUSED(webWindowType);
    if (WrtSettings::createWrtSettings()->value("PopupBlocking").toInt())
        return 0;
    WrtPage* wrtPage = new WrtPage();
    emit createWindow(wrtPage);
    return wrtPage;
}

void WrtPage::enableLocalStorage(const QString & path) {
    settings()->setAttribute(QWebSettings::LocalStorageEnabled, true);
    settings()->setLocalStoragePath(path);
 }

qint64 WrtPage::enableOfflineWebApplicationCache(const QString & path, const qint64 & quota) {
    if (!QFile::exists(path)) {
        QDir dir(path);
        if (!dir.mkdir(path)) {
            qCritical()<<"FAIL: Can't create missing application cache directory" << path;
            return -1;
        }
    }
    settings()->setAttribute(QWebSettings::OfflineWebApplicationCacheEnabled, true);
    settings()->setOfflineWebApplicationCachePath(path);
    if (quota > 0) { // If not set, it is by default OFFLINE_WEBAPP_DEFAULT_QUOTA
        settings()->setOfflineWebApplicationCacheQuota(quota);
    } else {
        settings()->setOfflineWebApplicationCacheQuota(OFFLINE_WEBAPP_DEFAULT_QUOTA);
    }

    return settings()->offlineWebApplicationCacheQuota();
}

void WrtPage::enablePersistentStorage(const QString& path)
{
    settings()->enablePersistentStorage(path);

    // NOTE: What about OfflineStorageDefaultQuota?
    if (settings()->offlineWebApplicationCacheQuota() == 0)
        settings()->setOfflineWebApplicationCacheQuota(OFFLINE_WEBAPP_DEFAULT_QUOTA);
}

#ifdef QTWRT_ENABLE_NOTIFICATIONS
void WrtPage::requestPermission(QWebFrame* frame, QWebPage::PermissionDomain domain)
{
    setUserPermission(frame, domain, PermissionGranted);
}

void WrtPage::checkPermission(QWebFrame* frame, QWebPage::PermissionDomain domain, QWebPage::PermissionPolicy& policy)
{
    switch (domain) {
    case NotificationsPermissionDomain:
        policy = PermissionGranted;
        break;
    default:
        break;
    }
}

void WrtPage::cancelRequestsForPermission(QWebFrame*, QWebPage::PermissionDomain)
{
    // Not supported yet
}
#endif // QTWRT_ENABLE_NOTIFICATIONS

/*!
// implementation of javaScriptConsoleMessage defined in WebKit qwebpage.h
// User settings are provided through WrtSettingsUI.
// The following choices are offered:
// Logging Off;Log to file;Show pop-up notes;Log to file and show pop-up notes
*/
void WrtPage::javaScriptConsoleMessage(const QString& messageSource, const QString& messageLevel, const QString& message, int lineNumber, const QString& sourceID)
{
    Q_UNUSED(messageSource);
    Q_UNUSED(messageLevel);
    javaScriptConsoleMessage(message, lineNumber, sourceID);
}

void WrtPage::javaScriptConsoleMessage(const QString& message, int lineNumber, const QString& sourceID)
{

    int jsSetting = WrtSettings::createWrtSettings()->value ( "JavaScriptConsoleLog" ).toInt();
    QString logFileName = WrtSettings::createWrtSettings()->value ( "JavaScriptConsoleLogFilePath" ).toString();
    QString msg = ( "WRT Console Log: "+ ( QDate::currentDate() ).toString ( Qt::SystemLocaleShortDate ) +" "+ ( QTime::currentTime() ).toString ( "hh:mm:ss ap" ) + \
            "\nMessage: " + message + "\nSource: "+sourceID+" : "+QString::number(lineNumber) );

    if ( jsSetting == 2 || jsSetting == 3 ){
        //display this simple dialog box for now.
        QMessageBox::about(view(), "JavaScript Console", msg);
    }
    if ( jsSetting == 1 || jsSetting == 3 ){
    // message is logged to file
        QFile jsFile ( logFileName );
        if ( !jsFile.open ( QIODevice::Append ) ) return;

        QTextStream jsLog ( &jsFile );
        jsLog << msg << "\n";
        jsFile.close();
    }
    if ( jsSetting == 4 ) {
        qDebug() << msg;
    }
    // Emit message in case some view manager wants to implement a better GUI to display console messages
    emit jsConsoleMessage("messageSource: N/A in QT46", "messageLevel: N/A in QT46", message, lineNumber, sourceID);
}

void WrtPage::allowDialogs(bool allow)
{
    d->m_allowDialogs = allow;
}

void WrtPage::cancelDialog()
{
    d->m_dialogResult.m_dialogAccepted = false;
    d->m_dialogResult.m_promptText.clear();

    if (d->m_openDialog) {
        d->m_openDialog->close();
        d->m_openDialog = 0;
    }
    if (d->m_eventLoop && d->m_eventLoop->isRunning()) {
        d->m_eventLoop->quit();
    }
}

bool WrtPage::event(QEvent* event)
{
#if QT_VERSION >= QT_VERSION_CHECK(4, 7, 0)
#if QTWEBKIT_VERSION >= QTWEBKIT_VERSION_CHECK(2, 1, 0)
    switch (event->type()) {
    case QEvent::TouchBegin:
    case QEvent::TouchUpdate: {
        d->m_touchDefaultPrevented = QWebPage::event(event);
        return d->m_touchDefaultPrevented;
    }
    case QEvent::TouchEnd: {
        d->m_touchDefaultPrevented = false;
        return QWebPage::event(event);
    }

    default:
        return QWebPage::event(event);
    }
#else // QTWEBKIT_VERSION_CHECK(2, 1, 0)
    // QtWebKit older than 2.1
    return QWebPage::event(event);
#endif
#else // QT_VERSION_CHECK(4, 7, 0)
    // Qt older than 4.7
    return QWebPage::event(event);
#endif
}

bool WrtPage::isTouchDefaultPrevented()
{
    return d->m_touchDefaultPrevented;
}

#if defined(Q_OS_MAEMO5)
QString WrtPage::chooseFile(QWebFrame* /*parentFrame*/, const QString& /*suggestedFile*/)
{
    // Not supported if Q_OS_MAEMO5 defined
    return QString();
}

void WrtPage::javaScriptAlert(QWebFrame*, const QString& /*msg*/)
{
    // Not supported if Q_OS_MAEMO5 defined
}

bool WrtPage::javaScriptConfirm(QWebFrame*, const QString& /*msg*/)
{
    // Not supported if Q_OS_MAEMO5 defined
    return false;
}

bool WrtPage::javaScriptPrompt(QWebFrame*, const QString& /*msg*/, const QString& /*defaultValue*/, QString* /*result*/)
{
    // Not supported if Q_OS_MAEMO5 defined
    return false;
}

#else // neither Q_OS_MAEMO5 nor Q_OS_MAEMO6 nor Q_OS_SYMBIAN
QString WrtPage::chooseFile(QWebFrame* parentFrame, const QString& suggestedFile)
{
    return QWebPage::chooseFile(parentFrame, suggestedFile);
}

void WrtPage::javaScriptAlert(QWebFrame* frame, const QString& msg)
{
    QWebPage::javaScriptAlert(frame, msg);
}

bool WrtPage::javaScriptConfirm(QWebFrame* frame, const QString& msg)
{
    return QWebPage::javaScriptConfirm(frame, msg);
}

bool WrtPage::javaScriptPrompt(QWebFrame* frame, const QString& msg, const QString& defaultValue, QString* result)
{
     return QWebPage::javaScriptPrompt(frame, msg, defaultValue, result);
}
#endif

}   // end of namespace WRT
