#include <QMetaMethod>
#include "qasynchandler.h"
#include "qasynctask.h"
#include "qisysinfo.h"
#include "qsysinfotypes.h"
#include "qsysinfofactory.h"
#include "qsysinfoproperties.h"
#include "qrequestinfo.h"

QAsyncHandler::QAsyncHandler() : transactionId(0)
{
}

QAsyncHandler& QAsyncHandler::instance()
{
    static QAsyncHandler instance;
    return instance;
}

void QAsyncHandler::handleRequestCompleted( int error, int transactionId,const ChannelInformation& listInfo )
{
    emit requestCompleted( error,transactionId,listInfo );

    QRequestInfo info;
    info.type = TypeAsynchrounous;
    info.transId = transactionId;
    criticalSection.lock();
    SysInfoProperty channelId = ongoingRequests.key(info );
    ongoingRequests.remove( channelId );    
    criticalSection.unlock();
    
}
void QAsyncHandler::handleRequestStarted( int error, int channelId,const ChannelInformation& listInfo )
{
    criticalSection.lock();
    QRequestInfo info  = ongoingRequests.value(channelId);
    criticalSection.unlock();

    emit requestStarted( error,info.transId,listInfo );

}
int QAsyncHandler::submitRequest( SysInfoProperty channelId, const int transId )
{
    int errCode = ErrNoError;
    QRequestInfo reqInfo;

    QISysInfo* instance = QSysInfoFactory::factoryInstance().getSysInfoInstance( channelId );
    if( !instance )
    {
        errCode = ErrDataNotFound;
    }
    else
    {
        reqInfo.type = TypeAsynchrounous;
        reqInfo.transId = transId;

        QObject::connect( instance, 
                SIGNAL( requestCompleted ( int, int,const ChannelInformation&)),
                this, 
                SLOT(handleRequestCompleted( int, int,const ChannelInformation&)),
                Qt::UniqueConnection );

        errCode = instance->getChannelAsync( channelId, transId );
        if( errCode == ErrNoError)
        {
            criticalSection.lock();
            ongoingRequests.insert( channelId, reqInfo );
            criticalSection.unlock();
        }
    }

    return errCode;
}

int QAsyncHandler::startChannel( SysInfoProperty channelId,
                                 const ChannelInformation& triggerInfo,
                                 const int transId)
{
    int errCode = ErrNoError;
    QRequestInfo reqInfo;

    QISysInfo* instance = QSysInfoFactory::factoryInstance().getSysInfoInstance( channelId );

    if( !instance )
    {
        errCode = ErrNotSupported;
    }
    else
    {
        criticalSection.lock();
        bool res =  ongoingRequests.contains( channelId );
        criticalSection.unlock();

        if( res == true )
        {
            errCode = stopChannel( channelId );
        }

        if(errCode == ErrNoError)
        {
            reqInfo.type = TypeNotification;
            reqInfo.transId = transId;
            
            QObject::connect( instance, 
                    SIGNAL( requestStarted( int, int,const ChannelInformation&)),
                    this, 
                    SLOT(handleRequestStarted( int, int,const ChannelInformation&)),
                    Qt::UniqueConnection );

            errCode = instance->startChannel( channelId, triggerInfo, transId );
            if( errCode == ErrNoError)
            {
                criticalSection.lock();
                ongoingRequests.insert( channelId, reqInfo );
                criticalSection.unlock();
            }

        }
    }
    return errCode;
}

int QAsyncHandler::stopChannel( SysInfoProperty channelId) 
{
    int errCode = ErrNoError;
    QISysInfo* instance = QSysInfoFactory::factoryInstance().getSysInfoInstance( channelId );

    if( !instance )
    {
        errCode = ErrDataNotFound;
    }
    else
    {
        criticalSection.lock();
        bool res =  ongoingRequests.contains( channelId );
        criticalSection.unlock();

        if( res == false )
        {
            errCode = ErrDataNotFound;
        }
        else
        {
            errCode = instance->stopChannel( channelId );
            if ( errCode == ErrNoError)
            {
                criticalSection.lock();
                QMap<SysInfoProperty,QRequestInfo>::iterator i = ongoingRequests.find(channelId);
                while (i != ongoingRequests.end() && i.key() == channelId) 
                {
                    if(i.value().type == TypeNotification )
                        ongoingRequests.erase(i);
                }
                criticalSection.unlock();
            }
        }
    }
    return errCode;
}

int QAsyncHandler::cancel( const int transactionId) 
{
    int errCode = ErrNoError;
    const int invalidTransId = -1 ;
    QRequestInfo info;
    info.type = TypeAsynchrounous;
    info.transId = transactionId;
    
    criticalSection.lock();
    SysInfoProperty channelId = ongoingRequests.key(info , invalidTransId );
    if(channelId != invalidTransId )
        ongoingRequests.remove( channelId );    
    criticalSection.unlock();
    
    if(channelId == invalidTransId)
        errCode = ErrInvalidArgument;
    else
    {
        QISysInfo* instance = QSysInfoFactory::factoryInstance().getSysInfoInstance( channelId );
        errCode = instance->cancel( channelId, transactionId );
    }

    return errCode;
}
