/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "qcodecinterface.h"
#include <QStringList>
#include <QVariant>

//VideoCodecs
const QString Kmanufacturer ( "manufacturer" );
const QString Kidentifier ( "identifier" );
const QString KmaxBitrate ( "maxBitrate" );
const QString KisHWAccelerationEnabled ( "isHWAccelerationEnabled" );
const QString Kversion ( "version" );

CodecRetriever::CodecRetriever() : m_gstValid(false)
{
   init();
}

void CodecRetriever::init()
{
    GError *err = 0;
    m_gstValid = gst_init_check(0, 0, &err);  //init gstreamer: must be called before any gst-related functions
    if (err)
        g_error_free(err);
}

CodecRetriever::~CodecRetriever()
{
    gst_deinit();
}

QStringList CodecRetriever::getList(QLatin1String codecType)
{
    qDebug() << "getList";
    QStringList availableMimeTypes;

    if (!m_gstValid)
    {
        return availableMimeTypes;
    }


    // Iterate over all audio and video decoders and extract mime types from sink caps
    GList* factoryList = gst_registry_get_feature_list(gst_registry_get_default (), GST_TYPE_ELEMENT_FACTORY);
    for (GList* iter =/* g_list_first*/(factoryList) ; iter != NULL ; iter = iter->next/* = g_list_next(iter)*/) {
        GstPluginFeature *feature = GST_PLUGIN_FEATURE(iter->data);
        QString klass = gst_element_factory_get_klass(GST_ELEMENT_FACTORY(feature));

        if (klass == QLatin1String(codecType)) {
            const GList *static_templates;
            GstElementFactory *factory = GST_ELEMENT_FACTORY(feature);
            static_templates = gst_element_factory_get_static_pad_templates(factory);

            for (; static_templates != NULL ; static_templates = static_templates->next) {
                GstStaticPadTemplate *padTemplate = (GstStaticPadTemplate *) static_templates->data;
                if (padTemplate && padTemplate->direction == GST_PAD_SINK) {
                    GstCaps *caps = gst_static_pad_template_get_caps (padTemplate);

                    if (caps) {
                        const GstStructure* capsStruct = gst_caps_get_structure (caps, 0);
                        QString mime = QString::fromUtf8(gst_structure_get_name (capsStruct));
                        if (!availableMimeTypes.contains(mime))
                              availableMimeTypes.append(mime);
                              qDebug() << "mime: " << mime;
                    }
                }
            }
        }
    }
    g_list_free(factoryList);
    availableMimeTypes.sort();
    return availableMimeTypes;
}
VideoCodecList CodecRetriever::getListwithDetails(QLatin1String codecType)
{
    qDebug() << "getList";
    VideoCodecList availableMimeTypes;

    if (!m_gstValid)
    {
        return availableMimeTypes;
    }


    // Iterate over all audio and video decoders and extract mime types from sink caps
    GList* factoryList = gst_registry_get_feature_list(gst_registry_get_default (), GST_TYPE_ELEMENT_FACTORY);
    for (GList* iter =/* g_list_first*/(factoryList) ; iter != NULL ; iter = iter->next/* = g_list_next(iter)*/) {
        GstPluginFeature *feature = GST_PLUGIN_FEATURE(iter->data);
        QString klass = gst_element_factory_get_klass(GST_ELEMENT_FACTORY(feature));

        if (klass == QLatin1String(codecType)) {
            const GList *static_templates;
            GstElementFactory *factory = GST_ELEMENT_FACTORY(feature);
            static_templates = gst_element_factory_get_static_pad_templates(factory);

            for (; static_templates != NULL ; static_templates = static_templates->next) {
                GstStaticPadTemplate *padTemplate = (GstStaticPadTemplate *) static_templates->data;
                if (padTemplate && padTemplate->direction == GST_PAD_SINK) {
                    GstCaps *caps = gst_static_pad_template_get_caps (padTemplate);

                    if (caps) {
                        const GstStructure* capsStruct = gst_caps_get_structure (caps, 0);
                        QString mime = QString::fromUtf8(gst_structure_get_name (capsStruct));
//                        if (!availableMimeTypes.contains(mime))
                        {
                              QMap <QString, QVariant> mimeDetails;
                              mimeDetails.clear();
//                            mimeDetails.insert(,iter->longname);

                              mimeDetails.insert(Kmanufacturer,gst_element_factory_get_author (GST_ELEMENT_FACTORY(feature)));
                              mimeDetails.insert(Kidentifier,mime);
                              mimeDetails.insert(KmaxBitrate,"");
                              mimeDetails.insert(KisHWAccelerationEnabled,"");
                              mimeDetails.insert(Kversion,"");

                              availableMimeTypes.append(mimeDetails);
                              qDebug() << "mime: " << mimeDetails;
                        }
                    }
                }
            }
        }
    }
    g_list_free(factoryList);
    return availableMimeTypes;
}

QVariant CodecRetriever::getListwithDetailsForSP(QLatin1String codecType)
{
   // VideoCodecList availableMimeTypes;
    QList< QVariant > availableMimeTypesForSP;

    if (!m_gstValid)
    {
        return static_cast<QVariant>(availableMimeTypesForSP);
    }


    // Iterate over all audio and video decoders and extract mime types from sink caps
    GList* factoryList = gst_registry_get_feature_list(gst_registry_get_default (), GST_TYPE_ELEMENT_FACTORY);
    for (GList* iter =/* g_list_first*/(factoryList) ; iter != NULL ; iter = iter->next/* = g_list_next(iter)*/) {
        GstPluginFeature *feature = GST_PLUGIN_FEATURE(iter->data);
        QString klass = gst_element_factory_get_klass(GST_ELEMENT_FACTORY(feature));

        if (klass == QLatin1String(codecType)) {
            const GList *static_templates;
            GstElementFactory *factory = GST_ELEMENT_FACTORY(feature);
            static_templates = gst_element_factory_get_static_pad_templates(factory);

            for (; static_templates != NULL ; static_templates = static_templates->next) {
                GstStaticPadTemplate *padTemplate = (GstStaticPadTemplate *) static_templates->data;
                if (padTemplate && padTemplate->direction == GST_PAD_SINK) {
                    GstCaps *caps = gst_static_pad_template_get_caps (padTemplate);

                    if (caps) {
                        const GstStructure* capsStruct = gst_caps_get_structure (caps, 0);
                        QString mime = QString::fromUtf8(gst_structure_get_name (capsStruct));
//                        if (!availableMimeTypes.contains(mime))
                        {

                              QMap <QString, QVariant> mimeDetails;
                              mimeDetails.clear();
//                            mimeDetails.insert(,iter->longname);

                              mimeDetails.insert(Kmanufacturer,gst_element_factory_get_author (GST_ELEMENT_FACTORY(feature)));
                              mimeDetails.insert(Kidentifier,mime);
                              mimeDetails.insert(KmaxBitrate,"");
                              mimeDetails.insert(KisHWAccelerationEnabled,"");

                              if (GST_PLUGIN_FEATURE (iter->data)->plugin_name) {
                                  GstPlugin *plugin;

                                  plugin = gst_registry_find_plugin (gst_registry_get_default (),
                                                                     GST_PLUGIN_FEATURE (iter->data)->plugin_name);
                                  if (plugin) {
                                      mimeDetails.insert(Kversion,plugin->desc.version);
                                  }else{
                                      mimeDetails.insert(Kversion,"N/A");
                                  }

                              }
                              availableMimeTypesForSP.append(static_cast<QVariant>(mimeDetails));
                              //availableMimeTypes.append(mimeDetails);
                              qDebug() << "mime: " << mimeDetails;
                        }
                    }
                }
            }
        }
    }
    g_list_free(factoryList);
    return static_cast<QVariant>(availableMimeTypesForSP);
}
