#include "qdeviceextensions_p.h"
#include "qsysinfoproperties.h"
#include "qsysinfoelement.h"
#include "qasynctask.h"


#include <QThreadPool>
#include <QDBusArgument>
#include <QDBusInterface>
#include <QDBusReply>
#include <QDBusMessage>
#include <QDBusConnection>

using namespace DeviceExtensions;
using namespace DeviceExtensions::ChannelCharging;
using namespace DeviceExtensions::ChannelBatteryLevel;
using namespace DeviceExtensions::ChannelBluetoothOn;
using namespace DeviceExtensions::ChannelDeviceInputMethod;
using namespace DeviceExtensions::ChannelVibrate;
using namespace DeviceExtensions::ChannelBacklightState;
using namespace DeviceExtensions::ChannelKeypadLightState;

#define HAL_BME_CHARGING                "battery.rechargeable.is_charging"

#define BLUEZ_SERVICE                   "org.bluez"
#define BLUEZ_PATH                      "/"
#define BLUEZ_IF_MANAGER                "org.bluez.Manager"
#define BLUEZ_IF_ADAPTER                "org.bluez.Adapter"
#define BLUEZ_GET_PATH                  "DefaultAdapter"
#define BLUEZ_GET_PROPERTY              "GetProperties"
#define BLUEZ_POWERED                   "Powered"
#define BLUEZ_CHANGE_PROPERTY           "PropertyChanged"

//Input Method
#define HAL_IM_SERVICE                  "org.freedesktop.Hal"
#define HAL_IM_PATH                     "/org/freedesktop/Hal/Manager"
#define HAL_IM_IF                       "org.freedesktop.Hal.Manager"
#define HAL_IM_FIND_DEVICE              "FindDeviceByCapability"
#define HAL_IM_KEYPAD                   "input.keyboard"
#define HAL_IM_TOUCHPAD                 "input.touchpad"

//MCE
#define MCE_SERVICE                     "com.nokia.mce"
#define MCE_REQUEST_PATH                "/com/nokia/mce/request"
#define MCE_REQUEST_IF                  "com.nokia.mce.request"
#define MCE_KEYPAD_GET_STATUS           "get_keyboard_status"
#define MCE_DISPLAY_GET_STATUS          "get_display_status"
#define MCE_START_VIBRATE               "req_start_manual_vibration"

#define HAL_BME_SERVICE                 "org.freedesktop.Hal"
#define HAL_BME_PATH                    "/org/freedesktop/Hal/devices/bme"
#define HAL_BME_IF                      "org.freedesktop.Hal.Device"
#define HAL_BME_GET_PROPERTY            "GetProperty"
#define HAL_BME_CHARGE_LEVEL            "battery.charge_level.percentage"
#define HAL_BME_CHANGE_PROPERTY         "PropertyModified"

QDeviceExtensionsPrivate::QDeviceExtensionsPrivate()
{
}

QDeviceExtensionsPrivate::~QDeviceExtensionsPrivate()
{
}

int QDeviceExtensionsPrivate::getChannel( SysInfoProperty channelId , 
        const ChannelInformation& input ,
        QMap<SysInfoProperty, QVariant>& channelInfo )

{
    int errCode = ErrNoError ;

    switch( channelId )
    {
        case Charging:
            {
                QDBusInterface interface( HAL_BME_SERVICE,
                        HAL_BME_PATH,
                        HAL_BME_IF,
                        QDBusConnection::systemBus());
                if ( true == interface.isValid() )
                {
                    QDBusMessage msg = interface.call( HAL_BME_GET_PROPERTY,
                            HAL_BME_CHARGING );
                    if ( msg.type() == QDBusMessage::ReplyMessage &&
                            msg.arguments().count() >= 1 )
                    {
                        bool isCharging = msg.arguments().at(0).toBool();
                        channelInfo.insert( chargingStatus , isCharging);
                    }
                    else
                        errCode =  ErrServiceNotAvailable;
                }
                else
                {
                    errCode =  ErrServiceNotAvailable;
                }
                break;
            }
        case BatteryLevel :
            {
                QDBusInterface interface( HAL_BME_SERVICE,
                        HAL_BME_PATH,
                        HAL_BME_IF,
                        QDBusConnection::systemBus());
                if ( true == interface.isValid() )
                {
                    QDBusMessage msg = interface.call( HAL_BME_GET_PROPERTY,
                            HAL_BME_CHARGE_LEVEL );
                    if ( msg.type() == QDBusMessage::ReplyMessage &&
                            msg.arguments().count() >= 1 &&
                            msg.arguments().at(0).type() == QVariant::Int )
                    {
                        qint32 percentage = msg.arguments().at(0).toInt();
                        channelInfo.insert( batteryLevel , percentage);
                    }
                    else
                        errCode =  ErrServiceNotAvailable;
                }
                else
                {
                    errCode =  ErrServiceNotAvailable;
                }
                break;
            }
        case BluetoothOn:
            {
                QDBusInterface interface( BLUEZ_SERVICE,
                        BLUEZ_PATH,
                        BLUEZ_IF_MANAGER,
                        QDBusConnection::systemBus());
                if ( true  == interface.isValid() )
                {
                    QDBusMessage msg = interface.call( BLUEZ_GET_PATH );
                    if ( msg.type() == QDBusMessage::ReplyMessage )
                    {
                        QString path = msg.arguments().at(0).value<QDBusObjectPath>().path();
                        QDBusInterface interface( BLUEZ_SERVICE,
                                path,
                                BLUEZ_IF_ADAPTER,
                                QDBusConnection::systemBus());
                        if ( true == interface.isValid() )
                        {
                            QDBusReply< QMap<QString,QVariant> > reply = 
                                interface.call( BLUEZ_GET_PROPERTY );
                            if ( reply.isValid() )
                            {
                                bool powered = reply.value().value( BLUEZ_POWERED ).toBool();
                                channelInfo.insert(btOn , powered );
                                break;
                            }
                            else
                            {
                                errCode =  ErrServiceNotAvailable;
                                break;
                            }
                        }
                        else
                        {
                            errCode =  ErrServiceNotAvailable;
                            break;
                        }
                    }
                }
                else
                    errCode =  ErrServiceNotAvailable;
                break;
            }
        case DeviceInputMethod:
            {
                bool keypad = FALSE;
                bool touchpad = FALSE;

                QDBusInterface interface( HAL_IM_SERVICE,
                        HAL_IM_PATH,
                        HAL_IM_IF,
                        QDBusConnection::systemBus());
                if ( FALSE == interface.isValid() )
                {
                    channelInfo.insert(deviceInputMethod,QString(""));
                }

                QDBusMessage msg_key = interface.call( QDBus::Block,
                        HAL_IM_FIND_DEVICE,
                        HAL_IM_KEYPAD );
                if ( msg_key.type() == QDBusMessage::ReplyMessage )
                {
                    QStringList list_key = msg_key.arguments().at(0).toStringList();
                    if ( FALSE == list_key.isEmpty())
                    {
                        keypad = TRUE;
                    }
                }
                else
                {
                    channelInfo.insert(deviceInputMethod,QString(""));
                }

                QDBusMessage msg_touch = interface.call( QDBus::Block,
                        HAL_IM_FIND_DEVICE,
                        HAL_IM_TOUCHPAD );
                if ( msg_touch.type() == QDBusMessage::ReplyMessage )
                {
                    QStringList list_touch = msg_touch.arguments().at(0).toStringList();
                    if ( FALSE == list_touch.isEmpty())
                    {
                        touchpad = TRUE;
                    }
                }
                else
                {
                    channelInfo.insert(deviceInputMethod,QString(""));
                }

                if ( TRUE == touchpad && TRUE == keypad  )
                {
                    channelInfo.insert(deviceInputMethod,"Hybrid");
                }
                if ( TRUE == touchpad && FALSE == keypad  )
                {
                    channelInfo.insert(deviceInputMethod,"Touch");
                }
                if ( FALSE == touchpad && TRUE == keypad  )
                {
                    channelInfo.insert(deviceInputMethod,"NonTouch");
                }
                break;
            }
        case BacklightState:
            {
                QDBusInterface interface( MCE_SERVICE,
                        MCE_REQUEST_PATH,
                        MCE_REQUEST_IF,
                        QDBusConnection::systemBus());
                if ( true == interface.isValid() )
                {
                    QDBusMessage msg = interface.call( MCE_DISPLAY_GET_STATUS );
                    if ( msg.type() == QDBusMessage::ReplyMessage )
                    {
                        QString status = msg.arguments().at(0).toString();
                        if ( "off" == status )
                        {
                            channelInfo.insert(backLightOn,false);
                        }
                        else
                        {
                            channelInfo.insert(backLightOn,true);
                        }
                        break;
                    }
                    else
                    {
                        errCode =  ErrServiceNotAvailable;
                    }
                }
                else
                {
                    errCode =  ErrServiceNotAvailable;
                }
                break;
            }
        case KeypadLightState:
	    {
            QDBusInterface interface( MCE_SERVICE,
                    MCE_REQUEST_PATH,
                    MCE_REQUEST_IF,
                    QDBusConnection::systemBus());
            if ( true == interface.isValid() )
            {
                QDBusMessage msg = interface.call(MCE_KEYPAD_GET_STATUS);

                if ( msg.type() == QDBusMessage::ReplyMessage &&  msg.arguments().count() >= 1 &&  msg.arguments().at(0).type() == QVariant::String )
                {
                    QString status = msg.arguments().at(0).toString();
                    if ( "off" == status )
                    {
                        channelInfo.insert(KeypadLightState,false);
                    }
                    else
                    {
                        channelInfo.insert(KeypadLightState,true);
                    }
                }
                else
                {
                    errCode = ErrServiceNotAvailable;
                }
            }
            else
            {
                errCode = ErrServiceNotAvailable;
            }
		    break;
	    }
        default:
        {
            errCode = QDeviceExtensions::getChannel(channelId, input, channelInfo);
            break;
        }

    }
    return errCode;
}

int QDeviceExtensionsPrivate::getChannelAsync( SysInfoProperty channelId , int transId)
{
    int errCode = ErrNoError;

    switch( channelId )
    {
        case BatteryLevel:
            {
                QAsyncTask* task = new QAsyncTask( transId, channelId );
                QThreadPool::globalInstance()->start( task );
                break;
            }
        default:
            errCode = QDeviceExtensions::getChannelAsync( channelId, transId ) ;
            break;
    }
    return errCode;
}

int QDeviceExtensionsPrivate::cancel( SysInfoProperty channelId , int transId)
{
    int errCode = ErrNoError; 
    switch( channelId )
    {
        default:
            errCode = QDeviceExtensions::cancel(channelId , transId);
    }
    return errCode ;
}

void QDeviceExtensionsPrivate::startBatteryLevelChannelCb(const QDBusMessage & message)
{
    int errCode = ErrNoError;
    if ( TRUE == triggerMap.contains( BatteryLevel ) )
    {
        QDBusArgument arg = message.arguments().at(1).value<QDBusArgument>();

        arg.beginArray();
        while ( !arg.atEnd() )
        {
            QStringList element;
            arg >> element;
            if ( HAL_BME_CHARGE_LEVEL == element.at(0) )
            {
                QMap<SysInfoProperty, QVariant> channelInfo;
                QMap<SysInfoProperty, QVariant> input;
                errCode = getChannel(BatteryLevel,input,channelInfo);

                if ( errCode == ErrNoError) 
                {
                    if ( (TRUE == triggerMap.contains(batteryLevel) ) )
                    {
                        ChannelInformation triggerInfo = triggerMap.value(batteryLevel);
                        QMap<SysInfoProperty,QVariant>::const_iterator i;
                        for( i = triggerInfo.constBegin(); i != triggerInfo.constEnd(); ++i)
                        {
                            SysInfoProperty prop =  i.key() ;
                            int threshold = i.value().toInt();
                            if( (prop == batteryLevel) || (prop == Default::ChanneldefaultKey::defaultKey) )
                            {
                                if(channelInfo.value(batteryLevel).toInt() > threshold)
                                    return;
                            }
                        }
                    }
                    emit requestStarted(errCode ,BatteryLevel , channelInfo); 
                }
            }
        }
        arg.endArray();
    }
}

void QDeviceExtensionsPrivate::startChargingChannelCb( const QDBusMessage & message )
{
    int errCode = ErrNoError;

    if ( TRUE == triggerMap.contains( Charging ) )
    {
        QDBusArgument arg = message.arguments().at(1).value<QDBusArgument>();
        arg.beginArray();
        while ( !arg.atEnd() )
        {
            QStringList element;
            arg >> element;
            if ( HAL_BME_CHARGING == element.at(0) )
            {
                QMap<SysInfoProperty, QVariant> channelInfo;
                QMap<SysInfoProperty, QVariant> input;

                errCode = getChannel(Charging,input,channelInfo);

                if ( errCode == ErrNoError) 
                {
                    emit requestStarted(errCode ,Charging , channelInfo); 
                }
            }
        }
        arg.endArray();
    }
    return;
}

void QDeviceExtensionsPrivate::startBluetoothOnChannelCb(const QDBusMessage & message)
{
    if ( TRUE == triggerMap.contains( BluetoothOn ) )
    {
        if ( BLUEZ_POWERED == message.arguments().at(0).toString() )
        {
            QDBusVariant dbusVariant = qvariant_cast<QDBusVariant>( message.arguments().at(1) );
            bool powered = dbusVariant.variant().toBool();
            ChannelInformation  channelInfo;
            channelInfo.insert(btOn, powered );
            emit requestStarted(ErrNoError,BluetoothOn , channelInfo); 
        }
    }
}
int QDeviceExtensionsPrivate::startChannel( SysInfoProperty channelId,
        const ChannelInformation& triggerInfo,
        int transId)
{
    int errCode = ErrNoError;
    switch(channelId)
    {
        case Charging:
            {
                if ( false == QDBusConnection::systemBus().connect( HAL_BME_SERVICE,
                            HAL_BME_PATH,
                            HAL_BME_IF,
                            HAL_BME_CHANGE_PROPERTY,
                            this,
                            SLOT(startChargingChannelCb(const QDBusMessage& ))) )
                {
                    errCode = ErrServiceNotAvailable;
                    break;
                }
                else
                    triggerMap.insert(channelId,triggerInfo);
                break;
            }
        case BatteryLevel:
            {
                if ( false == QDBusConnection::systemBus().connect( HAL_BME_SERVICE,
                            HAL_BME_PATH,
                            HAL_BME_IF,
                            HAL_BME_CHANGE_PROPERTY,
                            this,
                            SLOT(startBatteryLevelChannelCb(const QDBusMessage& ))) )
                {
                    errCode = ErrServiceNotAvailable;
                    break;
                }
                else
                {
                    QMap<SysInfoProperty,QVariant>::const_iterator i;
                    for( i = triggerInfo.constBegin(); i != triggerInfo.constEnd(); ++i)
                    {
                        SysInfoProperty prop =  i.key() ;
                        int threshold = i.value().toInt();
                        if( i.value().type() != QVariant::Double)
                        {
                            // null is a valid value to pass
                            //Passing null as input : i.value().type() = QVariant::Invalid ,  threshold = 0
                            if( (i.value().type() != QVariant::Invalid) && (threshold != 0) ) 
                            {
                                errCode = ErrInvalidArgument;
                                break;
                            }
                        }
                        if( (prop == batteryLevel) || (prop == Default::ChanneldefaultKey::defaultKey) )
                        {
                            if ( ( threshold < 0) || (threshold >= 100) )
                            {
                                errCode = ErrDataOutOfRange;
                                break;
                            }
                            triggerMap.insert(channelId,triggerInfo);
                            break;
                        }
                    }
                    
                }
                break;
            }
        case BluetoothOn:
            {
                QDBusInterface interface( BLUEZ_SERVICE,
                        BLUEZ_PATH,
                        BLUEZ_IF_MANAGER,
                        QDBusConnection::systemBus());
                if ( false == interface.isValid() )
                {
                    errCode = ErrServiceNotAvailable;
                    break;
                }

                QDBusMessage msg = interface.call( BLUEZ_GET_PATH );
                if ( msg.type() == QDBusMessage::ReplyMessage )
                {
                    QString path = msg.arguments().at(0).value<QDBusObjectPath>().path();
                    if ( false ==  QDBusConnection::systemBus().connect( BLUEZ_SERVICE ,
                                path,
                                BLUEZ_IF_ADAPTER,
                                BLUEZ_CHANGE_PROPERTY,
                                this,
                                SLOT(startBluetoothOnChannelCb(const QDBusMessage& ))) )
                    {
                        errCode = ErrServiceNotAvailable;
                        break;
                    }
                }
                else
                {
                    errCode = ErrServiceNotAvailable;
                    break;
                }
                
                break;
            }
        default:
            errCode = QDeviceExtensions::startChannel( channelId, triggerInfo, transId ) ;
    }
    triggerMap.insert(channelId,triggerInfo);
    return errCode;
}

int QDeviceExtensionsPrivate::stopChannel( SysInfoProperty channelId ) 
{
    int errCode = ErrNoError;
    switch( channelId)
    {
        case BatteryLevel:
            {
                QDBusConnection::systemBus().disconnect( HAL_BME_SERVICE,
                        HAL_BME_PATH,
                        HAL_BME_IF,
                        HAL_BME_CHANGE_PROPERTY,
                        this,
                        SLOT(startBatteryLevelChannelCb(const QDBusMessage& )));
                break;
            }
        case Charging:
            {
                QDBusConnection::systemBus().disconnect( HAL_BME_SERVICE,
                        HAL_BME_PATH,
                        HAL_BME_IF,
                        HAL_BME_CHANGE_PROPERTY ,
                        this,
                        SLOT(startChargingChannelCb(const QDBusMessage& )));
                triggerMap.remove(channelId);
                break;
            }
        case BluetoothOn:
            {
                QDBusInterface interface( BLUEZ_SERVICE,
                        BLUEZ_PATH,
                        BLUEZ_IF_MANAGER,
                        QDBusConnection::systemBus());
                if ( FALSE == interface.isValid() )
                {
                    errCode = ErrServiceNotAvailable;
                    break;
                }
                QDBusMessage msg = interface.call( BLUEZ_GET_PATH );
                if ( msg.type() == QDBusMessage::ReplyMessage )
                {
                    QString path = msg.arguments().at(0).value<QDBusObjectPath>().path();
                    QDBusConnection::systemBus().disconnect( BLUEZ_SERVICE ,
                            path,
                            BLUEZ_IF_ADAPTER,
                            BLUEZ_CHANGE_PROPERTY,
                    this,
                    SLOT(startBluetoothOnChannelCb(const QDBusMessage& )));
                }
                else
                {
                    errCode = ErrServiceNotAvailable;
                    break;
                }
                break;
            }
        default:
            errCode = QDeviceExtensions::stopChannel(channelId) ; 
    }

    return errCode;
}
int QDeviceExtensionsPrivate::setChannelSync( SysInfoProperty channelId ,const QVariant& input,QMap<SysInfoProperty, QVariant>& errData)
{
    int errCode = ErrNoError;

    switch(channelId)
    {
        case Vibrate:
            {
                if( input.type() != QVariant::Double )
                {
                    errCode = ErrInvalidArgument;
                    break;
                }
                QDBusInterface interface( MCE_SERVICE,
                        MCE_REQUEST_PATH,
                        MCE_REQUEST_IF,
                        QDBusConnection::systemBus());
                if ( FALSE == interface.isValid() )
                {
                    errCode = ErrServiceNotAvailable;
                    break;
                }

                qint32 duration = input.toInt();
                if(duration  < 0)
                {
                    duration = 0; //infinity
                }
                QDBusMessage msg = interface.call( MCE_START_VIBRATE, 50*2, duration);
                if ( msg.type() != QDBusMessage::ReplyMessage )
                {
                    errCode = ErrServiceNotAvailable;
                    break;
                }
               break;
            }
        default:
            errCode = QDeviceExtensions::setChannelSync( channelId ,input,errData);
            break;
    }
    return errCode;
}

int QDeviceExtensionsPrivate::getChannelList(QList<SysInfoProperty>& channelList) 
{
    channelList += BatteryLevel ;
    channelList += Charging ;
    channelList += BluetoothOn ;
    channelList += Vibrate ;
    channelList += DeviceInputMethod ;
    channelList += BacklightState ;
    channelList += KeypadLightState ;

    return 0;
}


