/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "qsysinfoservice.h"
#include <iostream>
#include <QList>
#include <QString>
#include "qkeystore.h"
#include "qisysinfo.h"
#include "qsysinfofactory.h"
#include "qasynchandler.h"
#include "qsysinfoelement.h"


const QString errCode("errCode");
const QString errMessage("errMessage");

// --------------------------------------------------------------------
// SysInfoService::SysInfoService()
// Constructor.
// --------------------------------------------------------------------
SysInfoService::SysInfoService(QObject* parent):QObject(parent)
{
    QAsyncHandler& handler = QAsyncHandler::instance();
    QObject::connect( &handler, 
            SIGNAL( requestCompleted ( int, int,const ChannelInformation&)),
            this, 
            SLOT(handleRequestCompleted( int, int,const ChannelInformation&)) ,
            Qt::DirectConnection);
    QObject::connect( &handler, 
            SIGNAL( requestStarted( int, int,const ChannelInformation&)),
            this, 
            SLOT(handleRequestStarted( int, int,const ChannelInformation&)));
}
// --------------------------------------------------------------------
// SysInfoService::~SysInfoService()
// Destructor.
// --------------------------------------------------------------------
SysInfoService::~SysInfoService()
{
}

// SysInfoService::setSecuritySession()
// Provides implementation for setSecuritySession derived
// from IServiceSecurity.
// --------------------------------------------------------------------
void SysInfoService::setSecuritySession( QtMobility::QAbstractSecuritySession *aSecSession )
{
    _sec_session = aSecSession;
}

QString SysInfoService::getErrorMessage( int err, const QMap<SysInfoProperty, QVariant>& errData )
{
    QString errDescription( QErrorStore::description( err ) );
    errDescription.trimmed();
    if( err )
    {
        QList<SysInfoProperty> errDatakeys = errData.keys(); 

        for( int i = 0 ; i < errDatakeys.size(); i++ ) 
        {
            SysInfoProperty prop = errDatakeys.at( i );
            errDescription += " " + QKeyStore::key( prop ) + 
                " " +  errData.value( prop ).toString() ;
        }
    }
    return errDescription.trimmed();
}

int SysInfoService::fromProperties( const QMap<SysInfoProperty, QVariant>& channelInfo, QMap<QString, QVariant>& channelInfoData )
{
    QMap<SysInfoProperty, QVariant>::const_iterator iter;
    for (iter = channelInfo.constBegin(); iter != channelInfo.constEnd(); ++iter)
    {
        QString key = QKeyStore::key( iter.key() );
        QVariant variant = iter.value();
        switch( variant.type() )
        {
            case QVariant::UserType:
                {
                    if( variant.canConvert< QList< QMap<SysInfoProperty, QVariant> > >() )

                    {
                        QList< QVariant > data;
                        QList< QMap<SysInfoProperty, QVariant> > channelInfoR = variant.value< QList< QMap<SysInfoProperty, QVariant> > >();
                        for( int i=0; i < channelInfoR.size(); i++ )
                        {
                            QMap<QString, QVariant> dataMap;
                            fromProperties( channelInfoR.at(i), dataMap );
                            data.append( dataMap );                           
                        }
                        channelInfoData.insert( key, data );
                    }
                    break;
                }
            default:
                {
                    channelInfoData.insert( key, variant );
                }
                
        }
    }
    return ErrNoError;
}


int SysInfoService::toProperties( const QMap<QString, QVariant>& channelInfoData, QMap<SysInfoProperty, QVariant>& channelInfo )
{
    QList<QString> keys = channelInfoData.keys();    
    for ( int i = 0; i < keys.size(); i++ )
    {
        QString key = keys.at(i);
        channelInfo.insert( QKeyStore::value(key),  channelInfoData.value(key) );
    }

    return ErrNoError;
}

// SysInfoService::setChannelSync()
// Sets the properties specified channel.
// This is a synchronous method.
// --------------------------------------------------------------------
QVariant SysInfoService::setChannelSync( const QString& aChannel, const QVariant& aInputData)
{
    QMap<QString,QVariant> returnData;
    ChannelInformation errData;

    int err = ErrNoError;
    
    if( aChannel.trimmed().size() == 0 )
    {
        err = ErrDataNotFound;
    }
    else
    {
        SysInfoProperty channelId = QKeyStore::value( aChannel );

        QISysInfo* instance = QSysInfoFactory::factoryInstance().getSysInfoInstance( channelId );
        if( instance )
        {
            err = instance->setChannelSync( channelId, aInputData, errData);
        }
        else 
        {
            err = ErrInvalidArgument;
            errData.insert( NULL, aChannel );
        }
    }

    returnData.insert( errCode, err );
    returnData.insert( errMessage, getErrorMessage( err, errData ) );
    
    return returnData;
}
// --------------------------------------------------------------------
// SysInfoService::getChannelSync()
// Get the properties of the specified channel.
// This is a synchronous method.
// --------------------------------------------------------------------
QVariant SysInfoService::getChannelSync( const QString& aChannel , const QMap<QString,QVariant>& aInputData )
{
    QMap<QString,QVariant> channelInfoData;
    QMap<SysInfoProperty, QVariant> channelInfo ; 
    int err = ErrNoError;

    if( aChannel.trimmed().size() == 0 )
    {
        err = ErrDataNotFound;
    }
    else
    {
        SysInfoProperty channelId = QKeyStore::value( aChannel );

        QISysInfo* instance = NULL; 
        QSysInfoFactory& factory =  QSysInfoFactory::factoryInstance();    
        instance =  factory.getSysInfoInstance( channelId ) ; 
        if( instance )
        {
            ChannelInformation inputInfo;
            toProperties( aInputData, inputInfo );

            err = instance->getChannel( channelId, inputInfo , channelInfo );
            if( !err )
            {
                fromProperties( channelInfo, channelInfoData );
            }
        }   
        else
        {
            err = ErrDataNotFound;
            channelInfo.insert( NULL, aChannel );
        }
    }
    channelInfoData.insert( errCode, err );
    channelInfoData.insert( errMessage, getErrorMessage( err, channelInfo ) );
    
    return channelInfoData;
}

// --------------------------------------------------------------------
// SysInfoService::getChannelSync()
// Gets the properties of the specified channel.
// This is a synchronous method.
// --------------------------------------------------------------------
QVariant SysInfoService::getChannelSync( const QString& aChannel )
{
    QMap<QString,QVariant> aInputData;
    return getChannelSync( aChannel, aInputData );
}

void SysInfoService::handleRequestCompleted( int error, int transactionId,const ChannelInformation& channelInfo )
{
    QMap<QString, QVariant> channelInfoData;

    fromProperties( channelInfo, channelInfoData );

    emit asyncCallback( error,transactionId, channelInfoData );
}

void SysInfoService::handleRequestStarted( int error, int transactionId,const ChannelInformation& channelInfo )
{
    QMap<QString, QVariant> channelInfoData;

    fromProperties( channelInfo, channelInfoData );

    emit asyncCallbackStChannel( error,transactionId, channelInfoData );
}
// --------------------------------------------------------------------
// SysInfoService::getChannelAsync()
// Gets the properties of the specified channel.
// This is an asynchronous method.
// --------------------------------------------------------------------
QVariant SysInfoService::getChannelAsync( const QString& aChannel , const int transactionId )
{
    QMap<QString,QVariant> returnData;

    int err = ErrNoError;
    
    if( aChannel.trimmed().size() == 0 )
    {
        err = ErrDataNotFound;
    }
    else
    {
        SysInfoProperty channelId = QKeyStore::value( aChannel );

        QAsyncHandler& handler = QAsyncHandler::instance();
        err = handler.submitRequest( channelId, transactionId );
        if( !err )
            returnData.insert( "transactionId", transactionId );
    }
    
    returnData.insert( errCode, err);
    returnData.insert( errMessage, "" );

    return returnData;
}


QVariant SysInfoService::cancel( const int aTransId )
{
    int err = QAsyncHandler::instance().cancel( aTransId );
    
    QMap<QString,QVariant> returnData;
    
    returnData.insert( errCode , err );
    returnData.insert( errMessage , "" );
    
    return returnData;
}

QVariant SysInfoService::startChannel( const QString& aChannel,const int transId, const QVariant& trigger )
{
    QMap<QString,QVariant> returnData;
    QMap<SysInfoProperty,QVariant> triggerInfo;

    int err = ErrNoError;
    
    if( aChannel.trimmed().size() == 0 )
    {
        err = ErrDataNotFound;
    }
    else
    {
        SysInfoProperty channelId = QKeyStore::value(aChannel);

        if( trigger.type() == QVariant::Map )
        {
            QMap<QString,QVariant> triggerMap = trigger.toMap();
            toProperties( triggerMap, triggerInfo );
        }
        else
            triggerInfo.insert( QKeyStore::value("defaultKey") , trigger );

        err = QAsyncHandler::instance().startChannel( channelId, triggerInfo, transId );
        if( !err )
            returnData.insert( "transactionId", transId );
        
    }

    returnData.insert( errCode, err );
    returnData.insert( errMessage, getErrorMessage( err, triggerInfo ) );

    return returnData;
}


// --------------------------------------------------------------------
// SysInfoService::stopChannel()
// Stops the notifications (started by startChannel) for the channel
// specified.
// --------------------------------------------------------------------
QVariant SysInfoService::stopChannel( const QString& aChannel )
{
    QMap<QString,QVariant> returnData;

    int err = ErrNoError;
    
    if( aChannel.trimmed().size() == 0 )
    {
        err = ErrDataNotFound;
    }
    else
    {
        SysInfoProperty channelId = QKeyStore::value(aChannel);
        err = QAsyncHandler::instance().stopChannel( channelId );
    }

    returnData.insert( errCode, err );
    returnData.insert( errMessage, "" );
    
    return returnData;
}

QVariant SysInfoService::getChannelList()
{
    int err = ErrNoError;
    QMap<QString,QVariant> returnData;

    QList<QISysInfo*> list = QSysInfoFactory::factoryInstance().getSysInfoInstances();
    QList<SysInfoProperty> channelList;
    QList<QVariant > schemaList;
    
    for( int i=0; i < list.size(); i++)
    {
        QISysInfo* inst = list.at( i );
        if( inst )
        {
            inst->getChannelList(channelList);
        }
    }

    for(int j=0; j< channelList.size(); j++)
    {
        QMap<QString,QVariant> schema;
        QKeyStore::propertySchema(channelList.at(j),schema);
        schemaList.append(schema);
    }
    
    return schemaList;
}

