/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <QDebug>
#include <QDir>
#include <QFileInfo>
#include <QFile>
#include <QDesktopServices>
#include <QDomDocument>
#include <QDateTime>
#include "CertificateManager.h"
#include "widgetmanagerconstants.h"
#include "wrtsettings.h"
#include "WidgetUtilsLogs.h"
#include <openssl/bio.h>
#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/ossl_typ.h>
#include <openssl/objects.h>

namespace WRT {

    CertificateManager::CertificateManager()
    {
        OpenSSL_add_all_algorithms();
    }
    CertificateManager::~CertificateManager()
    {
    }

    bool CertificateManager::retrieveDeviceCertificate(QByteArray topLevel, QByteArray& cert)
    {

        QString aki = getAuthorityKeyIdentifier(topLevel);

        // hard-coded path to public cert store for linux
        QString widgetCertPath = "";
        QString dataLocation =  QDesktopServices::storageLocation(QDesktopServices::DataLocation);
        widgetCertPath = QDir::toNativeSeparators(dataLocation);
        widgetCertPath = widgetCertPath+QDir::separator()+WIDGET_FOLDER;
        widgetCertPath = QDir::toNativeSeparators(widgetCertPath);
        if (! createDir(widgetCertPath)) {
            LOG( "CertificateManger::retrieveDeviceCertificate  - Could not create certificate path dir");
            return false;
        }
        widgetCertPath = widgetCertPath+QDir::separator()+"widgetCertificates";
        widgetCertPath = QDir::toNativeSeparators(widgetCertPath);
        if (! createDir(widgetCertPath)) {
            LOG( "CertificateManger::retrieveDeviceCertificate - Could not create certificate path dir" );
            return false;
        }

        // go through all certificates in widgetCertPath
        // the call to getSubjectKeyIdentifier assumes cert will be in .pem format
        QDir certDir = QDir(widgetCertPath);
        LOG( "widgetCertPath: " << widgetCertPath );
        QStringList certList = certDir.entryList(QDir::Files | QDir::NoDotAndDotDot);
        LOG( "certList size: " << certList.size());
        while (!certList.isEmpty()) {
            QString certPath = certList.takeFirst();
            certPath.prepend(widgetCertPath + QDir::separator());
            QFile f(certPath);
            LOG( "Testing certificate: "+ certPath << " Size= "<< f.size());
            f.open(QIODevice::ReadOnly);
            QByteArray certByteArray(f.readAll());

            QString ski = getSubjectKeyIdentifier(certByteArray);
            // pick a cert whose ski matches aki, convert to bytearray and return
            if (ski == aki) {
                cert = certByteArray;
                return true;
            }
        }

        return false;
    }

    /**
      * Takes a certificate from the widget's signature.xml x509 element, gets the AKI from the cert,
      * and chain validates the certificate stored locally against the trust root.
      *
      * Qt implementation does not perform chain validation - there is only 1 cert in the chain
      */
    bool CertificateManager::validateCertificates(QList<QByteArray> certs, QByteArray endEntity, QByteArray topLevel)
    {
        QByteArray rootCert;        
        if (!retrieveDeviceCertificate(topLevel, rootCert)) return false;

        rootCert = this->getSelfSignedRootCert(rootCert, certs);
        if (rootCert.isEmpty()) {
        LOG(  "could not find self-signed root certificate in cert store");
            return false;
        }
        // get AKI from OpenSSL
        QString aki = getAuthorityKeyIdentifier(topLevel);

        // retrieve these AKI values from the trustpolicyfile so we can modify trust domains after WRT ships
        WRT::WrtSettings* m_settings = WRT::WrtSettings::createWrtSettings();
        // get policy files from WrtSettings
        QString trustPath = m_settings->valueAsString("SecurityTrustPolicyFile");
        QFile trustfile(trustPath);

        if ( !trustfile.open(QIODevice::ReadOnly | QIODevice::Text) ) {
            LOG("Failed to open trustfile");
            return false;
        }

        QDomDocument trustxml;
        QByteArray trustcontent = trustfile.readAll();
        bool parsed = trustxml.setContent( trustcontent, true );
        if (!parsed) {
            LOG("could not parse trust policy file: " << trustPath);
            return false;
        }
        trustfile.close();

        QMap<QString,QString> extValues;
        QDomNodeList domains = trustxml.documentElement().elementsByTagName("domain");
        bool domainSearchDone = false;
        for (int i = 0; i< domains.count(); i++ ) {
              QDomNode domain = domains.item(i);
              QDomNodeList certificates = domain.toElement().elementsByTagName("certificate");
              for (int j = 0; j < certificates.count(); j++) {
                  QDomNode certificate = certificates.at(j);
                  QString certaki = certificate.attributes().namedItem("aki").nodeValue();
                  if (certaki == aki) {
                      // check appropriate x509 extensions
                      QDomNodeList extensions = certificate.toElement().elementsByTagName("extension");
                      for (int k=0; k < extensions.count(); k++) {
                          QDomNode extension = extensions.at(k);
                          QString name = extension.attributes().namedItem("name").nodeValue();
                          if (name.compare(KEY_USAGE, Qt::CaseInsensitive) == 0) {
                              extValues.insert(KEY_USAGE, extension.attributes().namedItem("value").nodeValue());
                          } else if (name.compare(EXT_KEY_USAGE, Qt::CaseInsensitive) == 0) {
                              extValues.insert(EXT_KEY_USAGE, extension.attributes().namedItem("value").nodeValue());
                          } else if (name.compare(CERT_POLICY, Qt::CaseInsensitive) == 0) {
                              extValues.insert(CERT_POLICY, extension.attributes().namedItem("value").nodeValue());
                          } // else ignore
                      }
                      domainSearchDone = true;
                      break;
                  }
              }
              if (domainSearchDone)
                  break;
        }

        // TODO: which certificate do we check for proper extensions?  Right now it's EE
        if (!checkExtensionValues(endEntity, extValues)) return false;

        // build the root certificate into an X509
        BIO* root_b = BIO_new_mem_buf(rootCert.data(),rootCert.size());
        X509* x509_rootCert = PEM_read_bio_X509(root_b, NULL, 0, NULL);
        BIO_vfree(root_b);

        // build the end entity cert to validate
        BIO* end_b =  BIO_new_mem_buf(endEntity.data(),endEntity.size());
        X509* x509_endEntity = PEM_read_bio_X509(end_b, NULL, 0, NULL);
        BIO_vfree(end_b);

        int error = 0;

        X509_STORE_CTX* ca_ctx = NULL;
        if (!(ca_ctx = X509_STORE_CTX_new())) {
            error = 1;
        }

        STACK_OF(X509)* untrustedChain = NULL;
        if (!(untrustedChain = sk_X509_new_null())) {
            LOG(  "memory allocation error for untrusted cert chain!");
            error = 2;
        }

        if (!error) {
            for (int i = 0; i < certs.length(); i++) {
                QByteArray cert = certs.at(i);
                BIO* b = BIO_new_mem_buf(cert.data(),cert.size());
                X509* untrusted_x509_link = PEM_read_bio_X509(b, NULL, 0, NULL);
                BIO_vfree(b);
                sk_X509_push(untrustedChain, untrusted_x509_link);
            }
        }

        if (!error) {
            if (X509_STORE_CTX_init(ca_ctx, NULL, x509_endEntity, untrustedChain) != 1) {
                error = 3;
            }
        }

        STACK_OF(X509)* trustedChain = NULL;
        if (!(trustedChain = sk_X509_new_null())) {
            LOG(  "memory allocation error for trusted cert chain!");
            error = 4;
        }
        if (!error) {
            sk_X509_push(trustedChain, x509_rootCert);
            X509_STORE_CTX_trusted_stack(ca_ctx, trustedChain);
        }

        if (!error) {
            int rc = X509_verify_cert(ca_ctx);
            if (rc != 1) {
                LOG(  "return code for verify cert: " << rc);
                int err = X509_STORE_CTX_get_error(ca_ctx);
                const char* errmsg = X509_verify_cert_error_string(err);
                LOG(  "error in X509_verify_cert: " << errmsg);
                error = 5;
            }
        }

        if (ca_ctx) X509_STORE_CTX_free(ca_ctx);
        if (untrustedChain) sk_X509_pop_free(untrustedChain, X509_free);
        if (trustedChain) sk_X509_pop_free(trustedChain, X509_free);
        if (x509_endEntity) X509_free(x509_endEntity);

        if (error) {
            LOG(  "Error:" << "CertificateManager::validateCertificate" << error);
            return false;
        }
        return true;
    }

    bool CertificateManager::buildCertMapping(QList<QByteArray> certs, QByteArray& endEntityCert, QByteArray& topLevel) {
        // associate AKI and SKI for each cert
        QMap<QString, QString> akiToski;

        // index a key to the cert that it came from
        QMap<QString, int> skiToCertNum;
        QMap<QString, int> akiToCertNum;


        // build a mapping of SKI to AKI
        for (int i = 0; i < certs.length(); i++) {
            QByteArray cert = certs.at(i);
            QString l_aki = getAuthorityKeyIdentifier(cert);
            QString l_ski = getSubjectKeyIdentifier(cert);
            if (l_aki.isEmpty() || l_aki == l_ski) {
                topLevel = cert;
                continue; // we don't need to include this cert in our mapping, we've already figured out that it's the root certificate
            }
            akiToski.insert(l_aki, l_ski);
            skiToCertNum.insert(l_ski, i);
            akiToCertNum.insert(l_aki, i);
        }

        QList<QString> certSKIs = akiToski.values();
        QList<QString> certAKIs = akiToski.keys();

        // find an AKI that does not match an SKI of a cert in the widget package
        if (topLevel.isEmpty()) {
            QMapIterator<QString, QString> i(akiToski);
            while (i.hasNext()) {
                i.next();
                if (! certSKIs.contains(i.key()) ) {
                    topLevel = certs[akiToCertNum.value(i.key())];
                    break;
                }
            }
        }

        // find a cert whose SKI is not referenced as the AKI of another cert
        for (int j = 0; j < certSKIs.length(); j++) {
            QString l_ski = certSKIs.at(j);
            if (! certAKIs.contains( l_ski ) ) {
                endEntityCert = certs[skiToCertNum.value(l_ski)];
                break;
            }
        }

        // if we have an AKI but not an end-entity, we're looking at a widget with a self-signed cert
        if (endEntityCert.isNull() && !topLevel.isEmpty()) {
            endEntityCert = certs[skiToCertNum.value(topLevel)];
        }

        if (!endEntityCert.isNull() && !topLevel.isNull() )
            return true;
        return false;
    }


    QString CertificateManager::getAuthorityKeyIdentifier(QByteArray certificate) {
        QString retVal = QString("");
        BIO * b = BIO_new_mem_buf(certificate.data(),certificate.size());
        X509 * cert = PEM_read_bio_X509(b, NULL, 0, NULL);
        if (!cert) {
            BIO_vfree(b);
            return retVal;
        }
        int extensions = X509_get_ext_count(cert);
        ASN1_OBJECT* aki_obj = OBJ_txt2obj(LN_authority_key_identifier,0);
        for (int i = 0; i < extensions; i++) {
            X509_EXTENSION *ext = X509_get_ext(cert, i);
            ASN1_OBJECT* ext_obj = ext->object;
            if (OBJ_cmp(aki_obj,ext_obj) == 0) {
                BIO *bio = BIO_new(BIO_s_mem());
                if (!X509V3_EXT_print(bio, ext, 0, 0))
                    M_ASN1_OCTET_STRING_print(bio,ext->value);
                BUF_MEM *buf;
                BIO_get_mem_ptr(bio, &buf);
                int bufLength = buf->length;

                retVal.resize(bufLength);
                for (int i = 0; i < bufLength; i++) {
                    retVal[i] = buf->data[i];
                }
                BIO_vfree(bio);
                break;
            }
        }
        ASN1_OBJECT_free(aki_obj);
        BIO_vfree(b);
        X509_free(cert);

        if (!retVal.isEmpty()) {
            int keyidloc = retVal.indexOf("keyid:");
            if (keyidloc != 0)
                return QString();
            int endkeyid = retVal.indexOf("\n");
            if (endkeyid != -1)
                retVal.truncate(endkeyid);
            retVal.remove(0,6); // 6 characters for "keyid:"
            retVal = retVal.simplified();
        }

        return retVal;
    }

    QString CertificateManager::getSubjectKeyIdentifier(QByteArray certificate) {
        QString retVal = QString("");
        BIO * b = BIO_new_mem_buf(certificate.data(),certificate.size());
        X509 * cert = PEM_read_bio_X509(b, NULL, 0, NULL);
        if (!cert) {
            BIO_vfree(b);
            return retVal;
        }
        int extensions = X509_get_ext_count(cert);
        ASN1_OBJECT* ski_obj = OBJ_txt2obj(LN_subject_key_identifier,0);
        for (int i = 0; i < extensions; i++) {
            X509_EXTENSION *ext = X509_get_ext(cert, i);
            ASN1_OBJECT* ext_obj = ext->object;
            if (OBJ_cmp(ski_obj,ext_obj) == 0) {
                BIO *bio = BIO_new(BIO_s_mem());
                if (!X509V3_EXT_print(bio, ext, 0, 0))
                    M_ASN1_OCTET_STRING_print(bio,ext->value);
                BUF_MEM *buf;
                BIO_get_mem_ptr(bio, &buf);
                int bufLength = buf->length;

                retVal.resize(bufLength);
                for (int i = 0; i < bufLength; i++) {
                    retVal[i] = buf->data[i];
                }
                BIO_vfree(bio);
                break;
            }
        }
        ASN1_OBJECT_free(ski_obj);
        BIO_vfree(b);
        X509_free(cert);

        return retVal;
    }


    /**
      * A method for checking the values of Mandatory x509 Extensions for the Nokia Online CA
      */
    bool CertificateManager::checkExtensionValues(QByteArray certificate, QMap<QString,QString>& expectedExtValues) {
        BIO * b = BIO_new_mem_buf(certificate.data(),certificate.size());
        X509 * cert = PEM_read_bio_X509(b, NULL, 0, NULL);
        if (!cert) {
            BIO_vfree(b);
            return false;
        }

        int extensions = X509_get_ext_count(cert);
        ASN1_OBJECT* keyUsage_obj = OBJ_txt2obj(LN_key_usage,0);
        ASN1_OBJECT* extendedKeyUsage_obj = OBJ_txt2obj(LN_ext_key_usage,0);
        ASN1_OBJECT* certPolicies_obj = OBJ_txt2obj(LN_certificate_policies,0);

        int extensionsFound = 0;
        for (int i = 0; i < extensions; i++) {
            X509_EXTENSION *ext = X509_get_ext(cert, i);
            ASN1_OBJECT* ext_obj = ext->object;

            BIO *bio;
            BUF_MEM *buf;
            QString val;
            if ( OBJ_cmp(keyUsage_obj,ext_obj) == 0 ||
                 OBJ_cmp(extendedKeyUsage_obj,ext_obj) == 0 ||
                 OBJ_cmp(certPolicies_obj,ext_obj) == 0 ) {
                bio = BIO_new(BIO_s_mem());
                if (!X509V3_EXT_print(bio, ext, 0, 0))
                    M_ASN1_OCTET_STRING_print(bio,ext->value);
                BIO_get_mem_ptr(bio, &buf);
                val = QString(buf->data);
                val.resize(buf->length);
                val = val.simplified();

                if ((OBJ_cmp(keyUsage_obj,ext_obj) == 0) && (expectedExtValues.value(KEY_USAGE) != "")) {
                    // is this: X509v3 Key Usage
                    if (val != expectedExtValues.value(KEY_USAGE)) {
                        LOG(  "Key Usage extension != 'digitalSignature'");
                        BIO_vfree(bio);
                        ASN1_OBJECT_free(ext_obj);
                        return false;
                    }
                    extensionsFound++;
                } else if ((OBJ_cmp(extendedKeyUsage_obj,ext_obj) == 0) && (expectedExtValues.value(EXT_KEY_USAGE) != "")) {
                    // is this: X509v3 Extended Key Usage
                    if (val != expectedExtValues.value(EXT_KEY_USAGE)) {
                        LOG(  "Extended Key Usage extension != expected result");
                        BIO_vfree(bio);
                        ASN1_OBJECT_free(ext_obj);
                        return false;
                    }
                    extensionsFound++;
                } else if ((OBJ_cmp(certPolicies_obj,ext_obj) == 0) && (expectedExtValues.value(CERT_POLICY) != "")) {
                    // is this: X509v3 Certificate Policies
                    if (val != expectedExtValues.value(CERT_POLICY)) {
                        LOG(  "Certificate Policies Policy Identifier != expected result");
                        BIO_vfree(bio);
                        ASN1_OBJECT_free(ext_obj);
                        return false;
                    }
                    extensionsFound++;
                }

                BIO_vfree(bio);
            }
            ASN1_OBJECT_free(ext_obj);
        }

        if (extensionsFound != expectedExtValues.size()) {
            LOG(  "All required extensions not found. Returning an error.");
            return false;
        }

        return true;
    }

    bool CertificateManager::createDir(const QString& path) {
        QDir dir(path);
        if (!dir.exists()){
            if (!dir.mkdir(dir.absolutePath())) {
                LOG(  "Could not create dir : " << path);
                return false;
            }
        }
        return true;
    }

    QByteArray CertificateManager::getSelfSignedRootCert(QByteArray certFromStore, QList<QByteArray>& certChain) {
        QString cert_aki = getAuthorityKeyIdentifier(certFromStore);
        QString cert_ski = getSubjectKeyIdentifier(certFromStore);

        // if both aki and ski exist, and they are not equal, we do not have a real root cert
        if ( cert_aki != cert_ski && !cert_aki.isEmpty() ) {
            QByteArray potentialRootCert;
            if ( retrieveDeviceCertificate(certFromStore, potentialRootCert) ) {
                certChain.append(certFromStore);
                return getSelfSignedRootCert(potentialRootCert, certChain);
            } else {
                return QByteArray();
            }
        } else {
            return certFromStore;
        }
    }

} // namespace WRT
