/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "DigSigService.h"
#include "signatureparser.h"

#include <QDebug>
#include <QStringList>
#include <QDir>
#include <QDirIterator>

DigSigService::DigSigService() {
    m_validCertificateAKI = QString("");
}

DigSigService::~DigSigService() {
}

WidgetValidationError DigSigService::validateWidget(QString path, bool cleanup) {
    QByteArray aCaCertFilePath;
    QStringList sigList;
    WidgetValidationError error;

    qDebug() << "isValidCert: finding signature:" << path;

    if (!findDigitalSignatures(sigList, path)) {
        if (cleanup)rmDir(path);
        return FindSignatureFailed;
    }

    // TODO: sort sigList by parsing out issuer and determining - currently it
    // is sorted to look at distributor signatures (in descending order) and
    // look at author-signature last, if necessary
    bool success = false;
    while (!sigList.isEmpty()) {
        QString signaturefile = QString(sigList.takeLast());
        qDebug() << "isValidCert: opening " << signaturefile;

        SignatureParser parser(path, signaturefile);

        qDebug() << "isValidCert: parsing...";
        if (!parser.parse()){
            qDebug() << "isValidCert: Digital signature parsing failed.";
            if (cleanup)rmDir(path);
            error = SignatureParseFailed;
            continue;
        }

        if (!parser.requiredElementsExist()){
            qDebug() << "isValidCert: error with required elements";
            if (cleanup)rmDir(path);
            error = SignatureElementsMissing;
            continue;
        }
        qDebug() << "isValidCert: passed requiredElementsExist";

        QByteArray encyptionCert;
        // TODO: add WidgetValidationError as arg
        if (!parser.validateCertificates(encyptionCert)){
            qDebug() << "isValidCert: Certificate validation failed.";
            if (cleanup)rmDir(path);
            error = ChainValidationFailed;
            continue;
        }
        qDebug() << "isValidCert: passed validateCertificate";

        QDir dir(path);
        dir.setFilter(QDir::NoDotAndDotDot | QDir::Files);
        QDirIterator dirItr(dir, QDirIterator::Subdirectories);
        QStringList files;
        while (dirItr.hasNext()) {
            dirItr.next();
            int basePathLen = dirItr.path().length();
            QString relativePath = dirItr.filePath();
            relativePath = relativePath.mid(basePathLen + 1);
            files.append(relativePath);
        }

        if (!parser.checkReferencesExist(files)){
            qDebug() << "isValidCert: Reference check failed";
            if (cleanup)rmDir(path);
            error = ReferencesMissing;
            continue;
        }
        qDebug() << "isValidCert: passed checkReferencesExist";

        // TODO: add WidgetValidationError as an arg
        if (!parser.validateSignature(signaturefile, encyptionCert)){
            qDebug() << "isValidCert: Signature validation failed.";
            if (cleanup)rmDir(path);
            error = SignatureDecryptionFailed;
            continue;
        }
        qDebug() << "isValidCert: passed validateSignature";

        if (!parser.validateReferences(files)){
            qDebug() << "isValidCert: Reference validation failed.";
            if (cleanup)rmDir(path);
            error = ReferenceValidationFailed;
            continue;
        }
        qDebug() << "isValidCert: passed validateReferences";

        success = true;
        m_validCertificateAKI = parser.getAKI();
        m_certificateAKIs = parser.getAKIs();
        break;
    }

    if (!success) return error;

    //All done, Valid widget";
    if (cleanup)rmDir(path);
    qDebug() << "isValidCert: Valid widget.";
    return ValidateOK;
}

QString DigSigService::getCertificateAKI() {
    return m_validCertificateAKI;
}

QStringList DigSigService::getCertificateAKIs() {
    return m_certificateAKIs;
}

//////
// PRIVATE FUNCTIONS
/////

bool DigSigService::findDigitalSignatures(QStringList& digSig, const QString& path)
{
    QDir dir;
    QMap<int,QString> sortMap;

    dir.setPath(path);

    qDebug() << "WgtWidget:findDigSig path:" << path;
    qDebug() << "findSignature: " << dir.absolutePath();

    QStringList entries = dir.entryList();

    QRegExp distSig("signature([1-9][0-9]?).xml");
    QString dirSlash =dir.path() + dir.separator();
    foreach(QString name, entries) {
        if (!name.compare("author-signature.xml")) {
            QString fullName = dirSlash + name;
            qDebug() << fullName;
            digSig.append(fullName);
        } else if (distSig.exactMatch(name) ) {
            QString fullName = dirSlash + name;
            qDebug() << fullName;
            int sigidx=distSig.cap(1).toInt();
            sortMap.insert(sigidx, fullName);
        }
    }
    for (QMap<int, QString>::const_iterator it=sortMap.constBegin();it !=sortMap.constEnd();++it)
        digSig.append(it.value());
    if (digSig.count() > 0)
        return true;
    qDebug() << "No signature.xml files" ;
    return false;
}

bool DigSigService::rmDir(const QString& path)
{
    // get all files and folders list in the current folder
    QDirIterator it(path, QDir::Files|QDir::AllDirs|QDir::Hidden|QDir::NoDotAndDotDot);
    QDir currentDir(path);

    while (it.hasNext()) {
        it.next();
        // get file/dir name
        QString cleanitem = QDir::cleanPath(it.fileInfo().absoluteFilePath());

        if ( it.fileInfo().isFile() ) {
            // delete file
            currentDir.remove(cleanitem);
        }
        else {
            // recursevly keep deleting the content
            rmDir(cleanitem);
            // now delete the empty folder
            currentDir.rmdir(cleanitem);
        }
    }
    return true;
}
