/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef WBENCHMARK_H
#define WBENCHMARK_H

#include <QDebug>
#include <QString>
#include <QTime>

namespace wperf {
class benchmarker {
    public:
        benchmarker(QString msg, uint iterations = 1) : mIterations(iterations), mCount(iterations), mMessage(msg) {
            // do at least one iteration
            if (mIterations == 0) {
                mIterations = 1;
            }
            // start the timer
            mTime.start();
            qDebug() << QString("====> %1").arg(msg);
        };

        ~benchmarker() {
            // get the elapsed time
            int t = mTime.elapsed();

            if (mCount == 1) {
                QTime delta(0, 0, 0, 0);
                delta = delta.addMSecs(t);
                qDebug() << QString("====> %1 took %2").arg(mMessage).arg(delta.toString("mm:ss.zzz"));
            }
            else if (mCount != 0) {
                // calculate an average based on the number of iterations
                t /= mCount;
                qDebug() << QString("====> %1 took an average of %2 milliseconds per iteration (total of %3 iterations)").arg(mMessage).arg(t).arg(mCount);
            }
        };

        uint iterate() {
            return mIterations--;
        }

    private:
        uint mIterations;
        uint mCount;
        QTime mTime;
        QString mMessage;
};

#ifdef ENABLE_WBENCHMARK_ALL

// prints a timestamp with given prefix
//
#define WTS(XXX) do { \
    QString __s(XXX); \
    QTime __ts = QTime::currentTime(); \
    qDebug() << __ts.toString("====> %1 hh:mm:ss.zzz").arg(__s); \
} while (0)


// prints time spent in milliseconds in current block of execution
//
#define WBM(XXX) \
    wperf::benchmarker __b(XXX)


// marks benchmarking start point with given identifier
//
#define WBM_BEGIN(XXX, XXXID) \
    wperf::benchmarker* __m4d##XXXID = new wperf::benchmarker(XXX, 1)

// prints the delta for given identifier
//
#define WBM_END(XXXID) \
    delete __m4d##XXXID

// empty definitions
#else
#define WTS(XXX)
#define WBM(XXX)
#define WBM_BEGIN(XXX, XXXID)
#define WBM_END(XXXID)
#endif // ENABLE_WBENCHMARK_ALL
} // namespace
#endif // WBENCHMARK_H
