/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <QHash>
#include <QList>
#include <QNetworkRequest>
#include <QNetworkReply>
#include <QNetworkProxy>
#include <QAuthenticator>
#include <QSslError>
#include <QDesktopServices>
#include <QDir>
#include <QCoreApplication>
#include <QWebFrame>

#include "wrtpage.h"
#include "wrtsettings.h"
#include "SuperWidget.h"
#include "webappinfo.h"
#include "wrtnetworkaccessmanager.h"
#include "wrtnetworkaccessmanager_p.h"
#include "wrtnetworkfilterlistener.h"
#include "wrtnetworkerrorreply.h"
#include "wrtnetworkreply.h"
#include "wrtwidgetreply.h"
#include "w3celement.h"
#include "secsession.h"
#include "SchemeHandler.h"
#include "dialogsprovider.h"

#ifndef Q_OS_SYMBIAN
// Symbian has no default cookie jar
// FIXME Other platforms should not have default jars.  They should
// explicitly manage the jar to ensure security.
#include "wrtcookiejar.h"
#endif

#ifdef CWRT_WIDGET_FILES_IN_SECURE_STORAGE
#include "storage.h"
#endif

// This allows you to whitelist access for a widget.
void Q_DECL_IMPORT qt_drt_whiteListAccessFromOrigin(const QString& sourceOrigin,
                                                 const QString& destinationProtocol,
                                                 const QString& destinationHost,
                                                 bool allowDestinationSubdomains);

namespace WRT {

const int maxHostRequestHeaderSize = 4096;

    // FIXME: this is a temp solution until webkit provides better API to keep track of
    // the content of current white list.  This white list might be out of sync with the
    // effective white list in webkit resulting in multiple adds to the webkit's white list
    //
    // TODO update this implementation when webkit provides API to retrieve the current
    // white list content
static QSet<QString> whiteList;

/*!
 *
 */
WrtNetworkAccessManagerPrivate::WrtNetworkAccessManagerPrivate(WidgetContainerBase* widgetContainerBase, QWidget* wrtView) :
 m_schemeHandler(new SchemeHandler())
, m_secSession(0)
, m_userAgent(0)
, m_qDiskCache(0)
, m_widgetContainerBase(widgetContainerBase)
, m_wrtPage(widgetContainerBase->wrtPage())
, m_wrtView(wrtView)
, m_widgetId(0)
#ifdef CWRT_WIDGET_FILES_IN_SECURE_STORAGE
, m_storage(Storage::createInstance(WIDGET_STORAGE))
#endif
{
}

/*!
 *
 */
WrtNetworkAccessManagerPrivate::~WrtNetworkAccessManagerPrivate()
{
    delete m_schemeHandler;
    delete m_qDiskCache;
#ifdef CWRT_WIDGET_FILES_IN_SECURE_STORAGE
    delete m_storage;
#endif
}

/*!
 *
 */
WrtNetworkAccessManager::WrtNetworkAccessManager(WidgetContainerBase* widgetContainerBase, QWidget* wrtView) : QNetworkAccessManager(widgetContainerBase->wrtPage()),
d(new WrtNetworkAccessManagerPrivate(widgetContainerBase, wrtView))
{
#ifndef Q_OS_SYMBIAN
    // Symbian has no default cookie jar
    // FIXME Other platforms should not have default jars.  They should
    // explicitly manage the jar to ensure security.
    CookieJar* j = new CookieJar(0, 0);
    if (j)
        QNetworkAccessManager::setCookieJar(j);
#endif

    setupCache();

    m_connected = false;
    connect(this, SIGNAL(authenticationRequired(QNetworkReply *, QAuthenticator *)), this, SLOT(slotAuthenticationRequired(QNetworkReply *, QAuthenticator *)));

    connect(this, SIGNAL(proxyAuthenticationRequired(const QNetworkProxy & , QAuthenticator * )), this, SLOT(slotProxyAuthenticationRequired(const QNetworkProxy &, QAuthenticator *)));

    // handles SSL errors such as expired certificates
    connect(this, SIGNAL(sslErrors(QNetworkReply*,QList<QSslError>)), this, SLOT(slotSslError(QNetworkReply*,QList<QSslError>)));
}

/*!
 *
 */
WrtNetworkAccessManager::~WrtNetworkAccessManager()
{
    delete d;

}

/*!
 *
 */
void WrtNetworkAccessManager::setSecuritySession(SecSession* secSession)
{
    d->m_secSession = secSession;
}

/*!
 *
 */
SecSession* WrtNetworkAccessManager::securitySession() const
{
    return d->m_secSession;
}

/*!
 * sets the host header for Http request
 * currently supports setting of accept headers only
 * @param name  :  Http request header name
 * @param value :  Http request header value to be set
 */
void WrtNetworkAccessManager::setHostRequestHeader(const QString& name, const QString& value)
{
    if (name.isNull()
        || name.isEmpty()
        || name.length() > maxHostRequestHeaderSize
        || value.length() > maxHostRequestHeaderSize)
        return;

    QString nameLowerCase = name.toLower();
    if (nameLowerCase == "accept" && d->m_hostRequestHeaders.contains(nameLowerCase)) {
        // accept header adds new values. Any other header replaces existing values
        if (!(value.isNull() || value.isEmpty())) {
            // rfc 2616 4.2 :
            // Multiple message-header fields with the same field-name MAY be
            // present in a message if and only if the entire field-value for that
            // header field is defined as a comma-separated list [i.e., #(values)].

            QString header = d->m_hostRequestHeaders.find(nameLowerCase).value() + " , " + value;
            if (header.length() <= maxHostRequestHeaderSize)
                d->m_hostRequestHeaders.insert(nameLowerCase, header);
        }
    } else if (nameLowerCase == "accept"
            || nameLowerCase == "user-agent"
            || nameLowerCase.startsWith("x-")) {
            d->m_hostRequestHeaders.insert(nameLowerCase, value);
    }
}

/*!
Public Slots:
    void slotAuthenticationRequired(QNetworkReply *, QAuthenticator *);
    void slotProxyAuthenticationRequired(const QNetworkProxy &, QAuthenticator *);
*/
/*!
 *
 */
void WrtNetworkAccessManager::slotAuthenticationRequired(QNetworkReply *reply, QAuthenticator *athenticator)
{
    QString username("");
    QString password("");

    if (d->m_wrtView != NULL) {
            if (DialogsProvider::getUsernamePassword(d->m_wrtView, username, password)) {
                    athenticator->setUser(username);
                    athenticator->setPassword(password);
        }

    }
}
/*!
 *
 */
void WrtNetworkAccessManager::slotProxyAuthenticationRequired(const QNetworkProxy & networkProxy, QAuthenticator* authenticator)
{
    QString username("");
    QString password("");

    if (d->m_wrtView != NULL) {
            if (DialogsProvider::getUsernamePassword(d->m_wrtView, username, password)) {
                    authenticator->setUser(username);
                    authenticator->setPassword(password);
        }

    }
}

void WrtNetworkAccessManager::slotSslError(QNetworkReply* reply, const QList<QSslError>& errors)
{
    qWarning() << Q_FUNC_INFO;
    // ignore SSL errors for now
    // TODO: define common policy for SSL errors
    reply->ignoreSslErrors();
}


/*!
 *
 */
QNetworkReply* WrtNetworkAccessManager::createRequest(Operation op, const QNetworkRequest &request, QIODevice *outgoingData)
{

    // Create reply
    QNetworkReply* reply = NULL;

    // Create non-const copy of request
    QNetworkRequest networkRequest = request;

    // Only symbian os needs connection management
    // Open connection, if request uses http-protocol
    if (!networkRequest.url().scheme().compare("http", Qt::CaseInsensitive) ||
        !networkRequest.url().scheme().compare("https", Qt::CaseInsensitive)){

        // If security session exits, check capabilities.
        if (d->m_secSession) {
            QStringList capability("Network");
            if (!d->m_secSession->isAllowed(capability)) {
                reply = new WrtNetworkErrorReply(QNetworkReply::ContentAccessDenied, "Not allowed", networkRequest.url());
                //a finished signal will be sent and the member is invoked as soon as the application enters the main event loop
                QMetaObject::invokeMethod(reply, "finished", Qt::QueuedConnection);
                return reply;
            }

        }
    }

    // Modify networkRequest
    QHash<QString, QString>::iterator it;
    for (it = d->m_hostRequestHeaders.begin(); it != d->m_hostRequestHeaders.end(); it++) {
        if (it.key().toLower().compare("accept")==0) {
            QString accept = networkRequest.rawHeader("accept");
            accept += " , ";
            accept += QString(it.value());
            networkRequest.setRawHeader("accept", accept.toUtf8());
        }
        else
            networkRequest.setRawHeader(it.key().toUtf8(), it.value().toUtf8());
    }

    // Prevent access to other widget's data
    if (((!networkRequest.url().scheme().compare("file", Qt::CaseInsensitive) ||
         !networkRequest.url().scheme().compare("widget", Qt::CaseInsensitive) ||
#if ENABLE(SANDBOX_SCHEMA)
         !networkRequest.url().scheme().compare("sandbox", Qt::CaseInsensitive))
#endif
        && fileAccessBlocked(networkRequest.url().path())) ||
        (networkRequest.url().toString().startsWith(d->m_appPath, Qt::CaseInsensitive)) &&
        fileAccessBlocked(networkRequest.url().toString())) {
        reply = new WrtNetworkErrorReply(QNetworkReply::ContentAccessDenied, "Not allowed", networkRequest.url());
        QMetaObject::invokeMethod(reply, "finished", Qt::QueuedConnection);
        return reply;
    }

    // Handle file scheme too for now
    if (!networkRequest.url().scheme().compare("file", Qt::CaseInsensitive)){
        // always assume get request
        WrtWidgetReply* reply = new WrtWidgetReply(networkRequest,
#ifdef CWRT_WIDGET_FILES_IN_SECURE_STORAGE
                                                   d->m_storage
#else
                                                   0
#endif
                                                   ,"file",getWidgetRoot());
        return reply;
    }

    // Handle widget scheme
    if (!networkRequest.url().scheme().compare("widget", Qt::CaseInsensitive) && widgetAccessAllowed(request)){
        enableCrossOriginXHR(networkRequest.url());
        // always assume get request
        WrtWidgetReply* reply = new WrtWidgetReply(networkRequest,
#ifdef CWRT_WIDGET_FILES_IN_SECURE_STORAGE
                                                   d->m_storage
#else
                                                   0
#endif
                                                   ,"widget", d->m_appPath);
        return reply;
    }

#if ENABLE(SANDBOX_SCHEMA)
    // Handle sandbox scheme
    if (!networkRequest.url().scheme().compare("sandbox", Qt::CaseInsensitive)){
        // always assume get request
        WrtWidgetReply* reply = new WrtWidgetReply(networkRequest,
#ifdef CWRT_WIDGET_FILES_IN_SECURE_STORAGE
                                                   d->m_storage,
#else
                                                   0,
#endif
                                                   "sandbox");
        return reply;
    }
#endif

    // Handle lib scheme
    if (!networkRequest.url().scheme().compare("lib", Qt::CaseInsensitive)) {
        enableCrossOriginXHR(networkRequest.url());
        // always assume get request
        WrtWidgetReply* reply = new WrtWidgetReply(request,
#ifdef CWRT_WIDGET_FILES_IN_SECURE_STORAGE
                                                   d->m_storage,
#else
                                                   0,
#endif
                                                   "lib", d->m_appPath);
        return reply;
    }

    // Requests with "data" scheme can/must be served right away without queueing the request (see code below) even if no
    // network connection is available. Also no access policy checks are required as the request is not pointing to a resource
    // (that it might not be allowed to access, but instead contains the actual encoded data.
    if (!networkRequest.url().scheme().compare("data", Qt::CaseInsensitive))
        return QNetworkAccessManager::createRequest(op, request);

    // access policy check
    if (!accessAllowed(networkRequest)) {
        reply = new WrtNetworkErrorReply(QNetworkReply::ContentAccessDenied, "Not allowed", networkRequest.url());
        //a finished signal will be sent and the member is invoked as soon as the application enters the main event loop
        QMetaObject::invokeMethod(reply, "finished", Qt::QueuedConnection);
        return reply;
    }

    if (d->m_schemeHandler->HandleSpecialScheme(networkRequest.url())) {
        // handled in scheme handler - block network access
        QNetworkReply* reply = new WrtNetworkErrorReply(QNetworkReply::OperationCanceledError, "Scheme Handled", networkRequest.url());
        QMetaObject::invokeMethod(reply, "finished", Qt::QueuedConnection);
        return reply;
    }

    //custom network request preprocessing, if error return.
    if (preprocessNetworkRequest(op, networkRequest)) {
        reply = new WrtNetworkErrorReply(QNetworkReply::OperationCanceledError, "Pre-process failure", networkRequest.url());
        //a finished signal will be sent and the member is invoked as soon as the application enters the main event loop
        QMetaObject::invokeMethod(reply, "finished", Qt::QueuedConnection);
        return reply;
    }

    // Enable HTTP pipelining  on outbound request
    networkRequest.setAttribute(QNetworkRequest::HttpPipeliningAllowedAttribute, true);

    // To improve content loading speed, enable unexpired HTTP cache items to be returned immediately if present
    // Prevents communication with server to check validity

    if ( (networkRequest.rawHeader(QByteArray("Cache-Control"))).toLower() == "max-age=0")
        networkRequest.setAttribute(QNetworkRequest::CacheLoadControlAttribute, QNetworkRequest::AlwaysNetwork);
    else
        networkRequest.setAttribute(QNetworkRequest::CacheLoadControlAttribute, QNetworkRequest::PreferCache);

    return QNetworkAccessManager::createRequest(op, networkRequest, outgoingData);
}

/*
 * hanndle Connection established signal
 * once connection is established, create real network replys from requests buffered in the request queue
 * setup the dummy reply created earlier to bridge the real network reply
 */

void WrtNetworkAccessManager::handleConnection()
{
    m_connected = true;
}

/*
 * network disconnected
 */
void WrtNetworkAccessManager::connectionError()
{
    m_connected = false;
}

/*!
 * Setup cache
 * Need to use WrtSettingsUI to setup Disk Cache Directory Path
 */
void WrtNetworkAccessManager::setupCache()
{

    if ( WrtSettings::createWrtSettings()->value ( "HTTPCacheEnabled" ).toBool() ) {

        QString path(WrtSettings::createWrtSettings()->value ( "HTTPCacheDirectory" ).toString());
        QDir diskCacheDir(path);

        //If fully-qualified path wasn't specified already via Settings, do platform-specific search & replace.
        if ( !diskCacheDir.isAbsolute() ) {
            diskCacheDir.setPath(fixupHTTPcachePath(path));
        }

        Q_ASSERT(diskCacheDir.isAbsolute());

        // NOTE: Each QNetworkAccessManager/WRTNetworkAccessManager object MUST get its own HTTP cache object ..
        // .. and dedicated cache directory! These are not shareable!

        delete d->m_qDiskCache;
#if defined(USE_PRIVATEHTTPCACHE)
        d->m_qDiskCache = new FeatherWeightCache(this);
#else
        d->m_qDiskCache = new QNetworkDiskCache(this);
#endif
        d->m_qDiskCache->setCacheDirectory(diskCacheDir.absolutePath());
        d->m_qDiskCache->setMaximumCacheSize(DISKCACHESIZE);

        setCache(d->m_qDiskCache);
    }

}

/*!
 *
 */
int WrtNetworkAccessManager::preprocessNetworkRequest(const QNetworkAccessManager::Operation &op, QNetworkRequest &request)
{
    int error = 0;
    QList<WRTNetworkFilterListener*>::iterator it;
    for (it = d->m_wrtNetworkFilterListener.begin(); it != d->m_wrtNetworkFilterListener.end(); it++) {
        error = (*it)->customNetworkRequestPreprocess(op, request);
        if (error)
          return error;
    }

    return error;
}

void WrtNetworkAccessManager::setBasePath(const QString& basePath)
{
    QString widgetPath = basePath;
    widgetPath.replace("\\","/");
    if (widgetPath.endsWith("/"))
        widgetPath.chop(1);
    int l = widgetPath.lastIndexOf("/");
    d->m_appPath = widgetPath.left(l + 1);// always at least one, +1 to include '/'
    bool ok;
    d->m_widgetId = widgetPath.right(widgetPath.length() - l - 1).toInt(&ok);
}
/* prevent access to other widget's data
 ie. file:///C:/private/200267c0/webapps/157925442/flip.bmp
 widget path = C:/private/200267c0/webapps/157925442/
 app path = C:/private/200267c0/webapps/
 because the file is under C:/private/200267c0/webapps it must be under 157925442 or a common dir
 ie.
 C:/data - ok
 C:/private/200267c0/webapps/ - ok
 C:/private/200267c0/webapps/157925442/ - ok
 C:/private/200267c0/webapps/lib/ - ok
 C:/private/200267c0/webapps/123456789/ - BLOCKED
*/
bool WrtNetworkAccessManager::fileAccessBlocked(const QString& urlPath)
{
    bool blocked(false);
    QString url = urlPath.toLower();

    int l = (d->m_appPath.isEmpty()) ? -1 : url.indexOf(d->m_appPath);
    if (l > -1) {
        bool appOk;
        int appIdStart = d->m_appPath.length() + l;
        int appIdEnd = url.indexOf("/",appIdStart);
        int appId = url.mid(appIdStart, appIdEnd - appIdStart).toInt(&appOk);
        if (appOk && appId != d->m_widgetId)
            blocked = true;
    }

    return blocked;
}


bool WrtNetworkAccessManager::accessAllowed(const QNetworkRequest& request)
{
// Allowing everything for Maemo5 while the bug #3690 is not closed.
// http://bugs.nokia-boston.com/bugzilla/show_bug.cgi?id=3690
#ifdef Q_OS_MAEMO5
   return true;
#endif

    QString widgetId(d->m_widgetContainerBase->widgetProperties()->id());

    WebAppInfo webAppInfo;
    if (!WebAppRegistry::instance()->isRegistered(widgetId, webAppInfo))
        return false;

    return true;
}

QString WrtNetworkAccessManager::getWidgetRoot(){
    if (!d->m_widgetContainerBase)
       return "";
    if (!d->m_widgetContainerBase->widgetProperties())
       return "";

    return d->m_widgetContainerBase->widgetProperties()->installPath();
}

/* Apply platform-specific rules to covert HTTP cache location into a fully
 * qualified directory path.
 */
QString WrtNetworkAccessManager::fixupHTTPcachePath(const QString& templateFormPath)
{
    QString fullPath(templateFormPath);

    QString baseDir (QDesktopServices::storageLocation(QDesktopServices::CacheLocation));

    //perform tag replacement
    fullPath.replace(QLatin1String("<toplevel>"), baseDir,  Qt::CaseInsensitive);

    return fullPath;
}

void WrtNetworkAccessManager::enableCrossOriginXHR(QUrl requestUrl)
{
    QString origin = (requestUrl.scheme() + "://" + requestUrl.host()).toLower();
    // TODO should retrieve the actual white list from webkit when the API
    // becomes available.  See header file for more details
    if (!(whiteList.contains(origin)))
    {
        whiteList << origin;
        addWhiteListEntriesForOrigin(origin);
    }
}

/*! Adds entries necessary to allow cross origin XHRs from origin to the webkit maintained whiteList.
 *  Right now the access is pretty open, but in the future should be restricted based on what the origin is.
 *  This should be done here if it is easy and not to many dependencies are introduced, in a derived class,
 *  or in an unrelated class employing the signal-slot mechanisms of Qt or a callback interface.
 *  If the derived class mechanism is not used this method should become private non-virtual maybe even inlined in
 *  enableCrossOriginXHR.
 */
void WrtNetworkAccessManager::addWhiteListEntriesForOrigin(QString origin)
{
    QString domain;
    qt_drt_whiteListAccessFromOrigin(origin, "http", domain, true);
    qt_drt_whiteListAccessFromOrigin(origin, "https", domain, true);
    qt_drt_whiteListAccessFromOrigin(origin, "lib", domain, true);
}

bool WrtNetworkAccessManager::widgetAccessAllowed(const QNetworkRequest &request)
{
    QObject* object = request.originatingObject();
    if (!object)
        return false;

    QWebFrame* originatingFrame = qobject_cast<QWebFrame*>(object);
    QUrl orgnUrl = originatingFrame->url();
    QUrl reqUrl = request.url();
    QString orgnHost = orgnUrl.host();
    QString reqHost = reqUrl.host();

    if (!orgnUrl.isEmpty() && (orgnHost != reqHost)){
         //If the iframe was created without content then the setting the src for it
        //would be rejected here. This is a workaround and proper solution should be
        //investigated. Fixes Maps loading
        if(!orgnUrl.toString().compare("about:blank") &&
           originatingFrame->parentFrame()){
        }else{
            return false;
        }
    }
    if (!m_allowedHosts.contains(reqHost)) {
        WebAppRegistry * webAppRegistry = WebAppRegistry::instance();
        if (webAppRegistry->isRegistered(reqHost)) {
            m_allowedHosts.append(reqHost);
        } else {
            return false;
        }
    }
    return true;
}

}
