/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "wrtglobal.h"

#include <QDateTime>
#include <QFileInfo>
#include <QDesktopServices>

#include "wrtwidgetreply.h"
#include "wrtsettings.h"
#include "SuperWidget.h"
#include "webapplocalizer.h"
#include "widgetmanagerconstants.h"
#include <QFile>
#include <QFileInfo>

#ifdef CWRT_WIDGET_FILES_IN_SECURE_STORAGE
#include "storage.h"
#endif

namespace WRT {


WrtWidgetReply::WrtWidgetReply(const QNetworkRequest& request,
                               Storage* storage,
                               const QString& protocol,
                               QString widgetRoot)
    : m_protocol(protocol.toLower()), QNetworkReply(), m_position(0)
#ifdef CWRT_WIDGET_FILES_IN_SECURE_STORAGE
    , m_storage(storage)
#endif
{
    Q_ASSERT(!request.url().scheme().compare(m_protocol, Qt::CaseInsensitive));
    QUrl url = request.url();
    QUrl newUrl;

    int libRootLength = 0;
    if (!m_protocol.compare("lib")) {
        // Format shared library url to absolute path so that
        // localization and secure storage checks can be done.

#if !defined(Q_OS_LINUX) && !defined(Q_OS_MEEGO)
        QString sharedLibPath(widgetRoot);
        QString sharedLibFolder("/"+SHARED_LIBRARY_FOLDER+"/");
        // when launching a widget with lib scheme the lib folder is already
        // part of the path.  this will prevent "/lib/" from being added twice
        if (!sharedLibPath.endsWith(sharedLibFolder))
            sharedLibPath.append(sharedLibFolder);
        QString libRoot("/" + QDir::fromNativeSeparators(sharedLibPath));
#else
        QString libRoot(QDir::fromNativeSeparators(SHARED_LIBRARY_PATH));
#endif
        libRootLength = libRoot.length(); // store root path length for faster conversion back to relative

        // Fix shared lib path if QUrl has parsed first folder of the path into authority field.
        // It does that if there are two slashes after scheme in url, like 'lib://'.
        if (!url.authority().isEmpty()) {
            url.setUrl("lib:" + libRoot + url.authority() + url.path());
        }
        else {
            url.setUrl("lib:" + libRoot + url.path().mid(1)); // remove leading '/' from path
        }
    }

    // Reconstruct the full path to the requested resource if this is a widget:// request
    if (protocol.compare("widget", Qt::CaseInsensitive) == 0)
    {
        // A widget scheme request is of form widget://<widget id>/<a path relative to the widget's private directory>
        // widgetRoot contains the private directory and the relative resource path can be retrieved with QUrl::path()
        // TODO: Do we want to see if the id in the requested url and the id in the widgetRoot matches?
        url = QUrl::fromLocalFile(widgetRoot + url.authority() + url.path());
    }

    if (getLocalizedUrl(newUrl,url))
    {
      url = newUrl;
    }

#ifdef CWRT_WIDGET_FILES_IN_SECURE_STORAGE
    // check if widget files are accessed from secure storage
    QString secFile = url.path();
#if !defined(Q_OS_LINUX) && !defined(Q_OS_MEEGO)
    if (secFile.startsWith("/"))
      {
    secFile = secFile.mid(1,-1);
      }
    secFile = QDir::toNativeSeparators(secFile);
#endif
    QFile secureFile(secFile);
    bool secure=false;
    if (m_storage) {
        secure = m_storage->verify(secureFile);
    }
#else
    bool secure = true; // no security/tamper checking
#endif

    m_fileName = toLocalFileName(url);

    if (!m_protocol.compare("lib")) {
        // Format library url back to relative by removing root path
        url.setUrl("lib://" + QDir::fromNativeSeparators(url.path()).mid(libRootLength));
    }

    // FIXME: should all reply URLs be the same as request URLs?
    if (!m_protocol.compare("widget"))
        setUrl(request.url());
    else
        setUrl(url);

    m_file.setFileName(m_fileName);
    bool exists = m_file.open(QIODevice::ReadOnly);

    if (exists) {
#ifdef CWRT_WIDGET_FILES_IN_SECURE_STORAGE
        // The code below is dead code if the above ifdef is not defined since secure is set to true
        if (!secure) {
            setAttribute(QNetworkRequest::HttpStatusCodeAttribute, 403);
            setAttribute(QNetworkRequest::HttpReasonPhraseAttribute, "Forbidden. Unsecure file access.");
            setError(ContentAccessDenied, "Access Denied");
            QMetaObject::invokeMethod(this, "sendError", Qt::QueuedConnection);
        }

        else {
#endif
            m_content = m_file.readAll();
            m_file.close();
            setAttribute(QNetworkRequest::HttpStatusCodeAttribute, 200);
            setAttribute(QNetworkRequest::HttpReasonPhraseAttribute, "ok");
            setHeader(QNetworkRequest::ContentLengthHeader, m_content.size());
            QFileInfo fileInfo(m_fileName);
            setHeader(QNetworkRequest::LastModifiedHeader, fileInfo.lastModified());
            QMetaObject::invokeMethod(this, "sendData", Qt::QueuedConnection);
            setOpenMode(QIODevice::ReadOnly);
#ifdef CWRT_WIDGET_FILES_IN_SECURE_STORAGE
        }
#endif
    }
    else {
        setAttribute(QNetworkRequest::HttpStatusCodeAttribute, 404);
        setAttribute(QNetworkRequest::HttpReasonPhraseAttribute, "Not found");
        setError(ContentNotFoundError, "Not found");
        QMetaObject::invokeMethod(this, "sendError", Qt::QueuedConnection);
    }
}

WrtWidgetReply::~WrtWidgetReply()
{
}
qint64 WrtWidgetReply::bytesAvailable () const
{
    return m_content.size();
}

QString WrtWidgetReply::toLocalFileName(QUrl& url)
{
    QString requestUrl = url.toString();

    // If the request was made with widget scheme, the url has already been reconstructed in the constructor
    // into a file URL with absolute path and can be treated the same way as a normal file request
    if (!m_protocol.compare("file") ||
        !m_protocol.compare("widget")) {
        return QUrl(requestUrl).toLocalFile();
    }

    if (!m_protocol.compare("lib")) {
#if !defined(Q_OS_LINUX) && !defined(Q_OS_MEEGO)
      return QDir::toNativeSeparators(url.path().mid(1)); // remove leading '/'
#else   // symbian || win32
      return QDir::toNativeSeparators(url.path());
#endif
    }

#if ENABLE(SANDBOX_SCHEMA)
    // For sandbox scheme
    int idEnd = requestUrl.indexOf("/", m_protocol.length() + 1); // length of "m_protocol:/"
    return QUrl(("file://" + requestUrl.mid(idEnd, -1))).toLocalFile();
#endif
}

void WrtWidgetReply::abort()
{
    setError(OperationCanceledError, ("Canceled"));
}

qint64 WrtWidgetReply::readData(char* data, qint64 size)
{
    qint64 len = qMin(size, (qint64)(m_content.size() - m_position));
    memcpy(data, m_content.data() + m_position, len);
    m_position += len;
    return len;
}

void WrtWidgetReply::sendData()
{
    QMetaObject::invokeMethod(this, "sendFinished", Qt::QueuedConnection);
    emit readyRead();
}

void WrtWidgetReply::sendFinished()
{
    emit finished();
}

void WrtWidgetReply::sendError()
{
    emit finished();
}

// function to get localized url from widget
// parameters:
//     newUrl     newUrl that represents localized content
//     reqUrl     requested url
// returns:
//     bool       true if successful
//
// This function is used only in Widget mode.
bool WrtWidgetReply::getLocalizedUrl(QUrl& newUrl, const QUrl& reqUrl)
{
    QString path = reqUrl.path();
    QStringList fileParts;

    // returns false if fileParts.count == 0, thus there's no need
    // to check the count of fileParts afterwards
    if (!SuperWidget::getWidgetPathParts(fileParts,path))return false;

    QString root = fileParts.at(0);
#if !defined(Q_OS_LINUX) && !defined(Q_OS_MEEGO)
    if (root.startsWith("/"))
      root = root.mid(1,-1);
#endif

    if (SuperWidget::getWidgetType(root)!=WidgetTypeW3c) return false;

#if defined(Q_OS_MAEMO5) || defined(Q_OS_MAEMO6) || defined(Q_OS_LINUX) || defined(Q_OS_MEEGO)
    QString locFile = WebAppLocalizer::getLocalizedFile(path, true, root);
#else   // symbian || win32
    QString locFile = WebAppLocalizer::getLocalizedFile(path.right(path.count()-1), true, root);
#endif

    if ( locFile.length() == 0 ) return false;
    locFile = "/"+locFile;

    newUrl.clear();
    newUrl.setPath(locFile);
    newUrl.setAuthority(reqUrl.authority());
    newUrl.setScheme(reqUrl.scheme());
    return true;
}
}
