/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "desktopfilereader.h"

#include <QFile>
#include <QTextStream>

namespace WRT {
    namespace Maemo {

        DesktopFileReader::DesktopFileReader(const QString & file) : m_valid(false), m_fileName(file) {
            QFile desktopFile(m_fileName);
            if (desktopFile.exists()) {
                m_valid = desktopFile.open(
                        QIODevice::ReadOnly|
                        QIODevice::Text);
                if (m_valid)
                {
                    QTextStream textStream(&desktopFile);
                    textStream.setCodec("UTF-8");

                    QString currentEntry;
                    while (!textStream.atEnd())
                    {
                        QString str = textStream.readLine();
                        if (!str.startsWith("#"))
                        {
                            if (str.startsWith("[") && str.endsWith("]"))
                            {
                                QMap<QString, QString> keys;
                                currentEntry = str.mid(1, str.size()-2);
                                m_desktopStructure.insert(currentEntry, keys);
                                continue;
                            }
                            if (str.contains("="))
                            {
                                QMap<QString, QString> keys = m_desktopStructure.value(currentEntry);
                                keys.insert(str.section("=", 0, 0).trimmed(), str.section("=", 1).trimmed());
                                m_desktopStructure.insert(currentEntry, keys);
                            }
                        }
                    }
                    desktopFile.close();
                }
            }
        }

        DesktopFileReader::~DesktopFileReader() {
            // Empty
        }

        bool DesktopFileReader::isValid() const {
            return m_valid;
        }

        void DesktopFileReader::validate() {
            // TODO : Move webwidget related stuff over here
            m_valid = true;
        }

        QString DesktopFileReader::fileName() const {
            return m_fileName;
        }

        QString DesktopFileReader::value(const QString & key, const QString & defaultValue) const {
            if (key.contains("/"))
            {
                return m_desktopStructure.value(key.section("/", 0, 0)).value(key.section("/", 1), defaultValue);
            }
            else
            {
                foreach (QString entryName, m_desktopStructure.keys())
                {
                    return m_desktopStructure.value(entryName).value(key, defaultValue);
                }
                return defaultValue;
            }
        }

        bool DesktopFileReader::contains(const QString & key) const {
            if (key.contains("/"))
            {
                return m_desktopStructure.value(key.section("/", 0, 0)).contains(key.section("/", 1));
            }
            else
            {
                foreach (QString entryName, m_desktopStructure.keys())
                {
                    if (m_desktopStructure.value(entryName).contains(key))
                    {
                        return true;
                    }
                }
                return false;
            }
        }

        QStringList DesktopFileReader::allKeys() const {
            QStringList retVal;
            foreach (QString entryName, m_desktopStructure.keys())
            {
                foreach (QString key, m_desktopStructure.value(entryName).keys())
                {
                    retVal.append(entryName + "/" + key);
                }
            }
            return retVal;
        }
    }

}
