/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <QDebug>
#include <QUrl>
#include <QStringList>

#include "WgtWidget.h"
#include "webapplocalizer.h"
#include "SuperWidget.h"
#include "webappinfo.h"
#include "webwidget.h"
#include "w3csettingskeys.h"
#include "wrtsettings.h"
#include "desktopfilereader.h"


namespace WRT {
    namespace Maemo {

        const QString TrustedDomain("TrustedWidgets");

        WebWidget::WebWidget(const WebAppInfo* webAppInfo,
                             QObject* parent) :
          QObject(parent)
          , m_webAppInfo(webAppInfo)
          , m_valid(true)
          , m_widget(0)
          , m_errorCode(NoError) {
            
            // This is only needed because of secure storage
            // For secure storage attribute it does not matter whether this
            // is /usr/share/webwidgets/applications/ .desktop file or
            // /usr/share/meegotouch/applets/ .desktop file
            // m_nativeId cannot be used reliably for getting viewmode
#if defined(Q_OS_MAEMO5) || defined(Q_OS_MAEMO6)
            DesktopFileReader tmpMetadata(m_webAppInfo->nativeId());
            m_valid = tmpMetadata.isValid();
            if (m_valid)
                init(tmpMetadata);
            else
                m_errorCode = MetadataError;
#else
            m_valid = true;
            // FIXME : Unused variable for linux build. Not wanted to change API. This is a hack for avoiding
            // file locked problem. On Fremantle DesktopFile is also loaded at WebWidgetRunner which lock the file (Raine).
            // From Frementle code path we should remove desktop-file use altogether.
            DesktopFileReader tmpMetadata("/tmp/foobar");
            init(tmpMetadata);
#endif
        }

        WebWidget::~WebWidget() {
            if (m_widget) {
                delete m_widget;
                m_widget = 0;
            }
        }

        bool WebWidget::isValid() {
            return m_valid;
        }

        WebWidget::WebWidgetErrorCode WebWidget::errorCode() {
            return m_errorCode;
        }

        QString WebWidget::getName() const {
            return m_properties.value("name").toString();
        }

        QUrl WebWidget::getUrl() const {
            return m_webAppInfo->url();
        }

        QString WebWidget::getWidgetResource() const {
            qDebug() << Q_FUNC_INFO << "Use WebAppInfo::appPath() instead of this function";
            return m_webAppInfo->appPath();
        }

        QString WebWidget::protectedStorageName() const {
            return m_properties.value("protected-store").toString();
        }
      
        QStringList WebWidget::supportedViewModes() const {
            qDebug() << Q_FUNC_INFO << "Use WebAppInfo::attributes().value(W3CSettingsKey::WIDGET_VIEWMODES).split(" ") instead of this function";
            QString viewmodes = m_webAppInfo->attributes().value(W3CSettingsKey::WIDGET_VIEWMODES).toString();           
            return viewmodes.split(" ");
        }

        QSize WebWidget::size() const {
            int width = 150;
            int height = 300;
            if (m_webAppInfo->attributes().contains(W3CSettingsKey::WIDGET_WIDTH))
                width = m_webAppInfo->attributes().value(W3CSettingsKey::WIDGET_WIDTH).toInt();
       
            if (m_webAppInfo->attributes().contains(W3CSettingsKey::WIDGET_HEIGHT))
                height = m_webAppInfo->attributes().value(W3CSettingsKey::WIDGET_HEIGHT).toInt();
           
            return QSize(width, height);
        }

      // FIXME: Remove WidgetProperties dependency from WidgetContainerBase
      // and then remove this function
      WidgetProperties* WebWidget::widgetProperties() {
          return ((SuperWidget*)m_widget)->getProperties();
      }


      void WebWidget::init(const DesktopFileReader & metadata) {

            // parses the name of the protected store
            // in case it is empty we assume non-signed widget
#if defined(Q_OS_MAEMO5) || defined(Q_OS_MAEMO6)
            m_properties.insert("protected-store", metadata.value("WRT/Protected-Store"));
#endif
            m_webAppId = m_webAppInfo->appId();
            m_widgetRoot = m_webAppInfo->appPath();

            m_widget = new WgtWidget(m_widgetRoot);
            m_widget->setWidgetUnZipPath(m_widgetRoot);
            m_widget->setWidgetInstallPath(m_widgetRoot);
            m_widget->m_validCertificateAKI = m_webAppInfo->certificateAki();
            
            if (!m_widget->parseManifest()) {
                qDebug()<<Q_FUNC_INFO<<"Error: ParseManifestFailed, line: "<<__LINE__;
                m_valid = false;
                m_errorCode = DesktopFileError;
                return;
            }

            m_widget->widgetProperties()->setId(m_webAppId);
            //parse manifest adds widget id to the end of install
            //so we need to set it here again. otherwise the installpath is /path/to/widgetA/widgetA
            //RAINE FIXME! should the WgtWidget be fixed !?
            m_widget->widgetProperties()->setInstallPath(m_widgetRoot);

#if defined(Q_OS_MAEMO5) || defined(Q_OS_MAEMO6)
            if (metadata.contains("WRT/SecureStore")) {
                m_properties.insert("secstore", metadata.value(
                        "WRT/SecureStore"));
            }
#endif

            WRT::WrtSettings* m_settings = WRT::WrtSettings::createWrtSettings();
            m_locale = m_settings->valueAsString("UserAgentLanguage");
            if (m_locale.isEmpty() || m_locale.compare(" ") == 0)
                m_locale = QLocale::system().name().toLower().replace(QString("_"),QString("-"));

            readConfigfile();
            //check Capabilities
            //Check from registry if policy files have been updated
            if (m_webAppInfo->capabilityCheck()) {
                WidgetFeatures features;
                WRT::SecSession* secureSession = 0;
                QString trustDn;
                QString sigId;
                bool res = m_widget->findFeatures(features);
                if (m_widget->createSecuritySession(&secureSession,trustDn,sigId,true)) {
                    if (res || (trustDn.compare(TrustedDomain,Qt::CaseInsensitive) == 0)) {
                        qDebug() << "Found features at runtime capability check" << features;
                        if (!m_widget->isFeatureAllowed(features,secureSession,true)) {
                            qDebug()<<Q_FUNC_INFO<<"Error: Feature is not allower, line: "<<__LINE__;
                            m_valid = false;
                            m_errorCode = CapabilityValidationError;

                        } else {
                            //capability check ok then save secsession file
                            QString resourcesDir = this->widgetProperties()->resourcePath();
                            QString secureSessionString = this->widgetProperties()->secureSessionString();
                            if (!resourcesDir.isNull() && !resourcesDir.isEmpty())
                               secureSession->saveSecsession(resourcesDir,secureSessionString);
                        }
                    }
                }
                //update the capabilityCheck Flag to false in WebappRegistry
                WebAppRegistry::instance()->setCapabilityCheck(m_webAppId,false);

                delete secureSession;
            }

            QString startFile = "";
            m_widget->findStartFile(startFile, m_widgetRoot);
            startFile = QString("widget://" + m_webAppId + "/" + startFile);
            m_properties.insert("path", startFile);
      }


        bool WebWidget::readConfigfile() {

            // FIXME : What if set language is invalid. We could try all the languages that we
            // get from languages list.

            // name
            if (m_widget->contains(W3CSettingsKey::WIDGET_NAME)) {
                m_properties.insert("name", m_widget->value(
                        W3CSettingsKey::WIDGET_NAME));
            }

            // icon

            if (this->widgetProperties() && !this->widgetProperties()->iconPath().isEmpty()) {
                QString icon = this->widgetProperties()->iconPath();
                m_properties.insert("icon", icon);
            }
            return true;
        }

    }

}
