/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "jsmenuitem.h"
#include "jsmenuitemfactory.h"
#include "action.h"

#include <QWebPage>
#include <QWebFrame>
#include <QFileInfo>
#include <QDebug>
#include <QIcon>


namespace WRT {

/*!
    \class WRT::JSMenuItem
    \brief The WRT::JSMenuItem represents the native instance of the MenuItem exposed to JavaScript.

    \inmodule WidgetCore

    WRT::JSMenuItem represents the native instance of the MenuItem expose to JavaScript. WRT::JSMenuItem holds
    a WRT::Action internally which is used as native menu item.
*/

/*!
    Constructs an instance of JSMenuItem with expose name \a exposeName
 */
JSMenuItem::JSMenuItem(QString exposeName)
    : JSResourceWrapper("", exposeName, JSResourceWrapper::DontEvalauteJavaScript)
    , m_action(new Action()) {
}

JSMenuItem::~JSMenuItem() {
    delete m_action;
}

/*!
    Gets \a name for the menu item.
*/
QString JSMenuItem::text() const {
    return m_action->text();
}

/*!
    Sets \a name for the menu item.
*/
void JSMenuItem::setText(QString name) {
    m_action->setText(name);
}

/*!
    Gets \a icon of the menu item.
*/
QString JSMenuItem::icon() const {
    return m_icon;
}

/*!
    Sets \a icon for the menu item. \a icon is a path from widget root to
    image file. E.g. if image.png is located in widget root (directly under
    install directory), \a icon is just image.png. If there is a localized
    icon in locales/en/ directory, then \a icon would be locales/en/image.png.

    In order to set icon succesfully, setWidgetId() and setAppPath must be
    called before calling setIcon().
*/
void JSMenuItem::setIcon(QString icon) {
    QUrl iconUrl(icon);
    if (!iconUrl.isValid() || !iconUrl.isRelative()) {
        qWarning() << Q_FUNC_INFO << "invalid icon: " << icon;
        return;
    }

    // TODO: handling of remote urls
    QString iconPath = iconUrl.path();

    if ( m_appPath.isEmpty() || m_widgetId.isEmpty()) {
        qWarning() << Q_FUNC_INFO << "m_appPath or m_widgetId is empty.";
        return;
    }

    // resolving absolute path to icon
    QString iconAbsolutePath = m_appPath + iconPath;
    QFileInfo iconInfo(iconAbsolutePath);
    if (iconInfo.exists()) {
        // This needs to be tested. setIcon has only been tested fo m6.
        QIcon newIcon(iconAbsolutePath);
        m_action->setIcon(newIcon);
        m_action->setIconVisibleInMenu(true);
        m_icon = icon;
    }
}

/*!
    Gets \a disabled from native instance.
*/
bool JSMenuItem::isDisabled() const {
    return !m_action->isEnabled();
}

/*!
    Sets native instance as \a disabled.
*/
void JSMenuItem::setDisabled(bool disabled) {
    m_action->setEnabled(!disabled);
}

/*!
    Returns a map (JS object, when called from JavaScript), which contains 
    key "supported". If "supported" is true, calling this function from JS is
    supported and the map contains also key "activated".
*/
QVariantMap JSMenuItem::isActivated() {
    QVariantMap retMap;

#ifdef QTWRT_ENABLE_TABMODE_TOOLBAR
    retMap.insert("supported", true);
    bool activated;
    // action is checkable if and only if action is in tab mode toolbar
    // isChecked tells the state in tab mode toolbar
    // isActivated tells the state when not in toolbar or in action mode toolbar
    if (m_action->isCheckable())
        activated = m_action->isChecked();
    else
        activated = m_action->isActivated();

    retMap.insert("activated", activated);
#else
    retMap.insert("supported", false);
#endif
    return retMap;
}

/*!
    Sets native instance as \a activated.
*/
void JSMenuItem::setActivated(bool activated) {
#ifdef QTWRT_ENABLE_TABMODE_TOOLBAR
    m_action->setActivated(activated);

    // check if menuitem is in tab mode toolbar
    if (m_action->isCheckable()) {
        // if activated == false, just calling setChecked(false) does not
        // seem to be enough to "uncheck" the action. setCheckable(false) is
        // also required.
        m_action->setCheckable(activated);
        m_action->setChecked(activated);
        m_action->setCheckable(true);
    }
#endif
}

/*!
    Sets the web application path.
*/
  void JSMenuItem::setAppPath(QString path) {
      m_appPath = path;
  }

/*!
    Sets the web application path.
*/
void JSMenuItem::setWidgetId(QString widgetId) {
    m_widgetId = widgetId;
}

/*!
    Returns a pointer to the native action of this action.
*/
Action* JSMenuItem::action() {
    return m_action;
}


void JSMenuItem::addToJavaScriptWindowObject() {
    if (!webPage()) {
        return;
    }

    webPage()->mainFrame()->addToJavaScriptWindowObject(name(), this);
}

/*!
    Adds a menu item owned by \a nativeInstanceHandle.
*/
bool JSMenuItem::addMenuItem(int nativeInstanceHandle) {
    // Maemo6 support as of now only one level of menu items
#if defined(SUPPORT_ONLY_FLAT_MENUS)
    return false;
#else
    JSMenuItem* item = JSMenuItemFactory::item(nativeInstanceHandle);
    m_action->addSubAction(item->action());
    return true;
#endif
}

/*!
    Removes a menu item owned by \a nativeInstanceHandle as a sub menu item to this menu item.
*/
bool JSMenuItem::removeMenuItem(int nativeInstanceHandle) {
    // Maemo6 support as of now only one level of menu items
#if defined(SUPPORT_ONLY_FLAT_MENUS)
    return false;
#else
    JSMenuItem* item = JSMenuItemFactory::item(nativeInstanceHandle);
    m_action->removeSubAction(item->action());
    return true;
#endif
}

/*!
    Inserts a menu item owned by \a nativeInstanceHandle as a sub menu item to this menu item before given \a beforeIndex.
*/
bool JSMenuItem::insertMenuItem(int nativeInstanceHandle, int beforeIndex) {
    // Maemo6 support as of now only one level of menu items
#if defined(SUPPORT_ONLY_FLAT_MENUS)
    return false;
#else
    JSMenuItem* item = JSMenuItemFactory::item(nativeInstanceHandle);
    m_action->insertSubAction(item->action(), beforeIndex);
    return true;
#endif
}

}
