/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <QVariant>
#include <QFile>
#include <QObject>
#include <QSignalMapper>
#include <qwebframe.h>

#include "jstoolbar.h"
#include "jsmenuitem.h"
#include "jsmenuitemfactory.h"
#include "action.h"
#include "actionadapter.h"


#if defined(Q_OS_MAEMO5) || defined(Q_OS_MAEMO6)
    static const int MAX_ITEM_COUNT = 4;
#else
    // TODO: define max size for other platforms, if needed
    static const int MAX_ITEM_COUNT = 99;
#endif


namespace WRT {


/*!
    \class WRT::JSToolbar
    \brief The WRT::JSToolbar

    \inmodule WidgetCore

    WRT::JSToolbar evaluates menu.js JavaScript file into window object.
*/

/*!
    Constructs an instance of JSToolbar that evaluates menu.js resource into current frame. Native part is exposed to menuCPP.
*/
JSToolbar::JSToolbar()
    : JSMenu("toolbar.js", "toolbarCPP")
    , m_mode(ToolbarAction)
{
}

JSToolbar::~JSToolbar()
{
}

int JSToolbar::maxSize()
{
    return MAX_ITEM_COUNT;
}

/*!
    Gets \a mode of the escape button. Returns a map (JS object, when called
    from JavaScript), which contains key "supported". If "supported" is true,
    calling this function from JS is supported and the map contains also
    key "mode".
*/
QVariantMap JSToolbar::mode() const
{
    QVariantMap retMap;

#ifdef QTWRT_ENABLE_TABMODE_TOOLBAR
    retMap.insert("supported", true);

    QString modeStr;
    switch (m_mode) {
    case ToolbarAction: 
      modeStr = "action";
      break;
    case ToolbarTab:
      modeStr = "tab";
      break;
    default:
      modeStr = "action"; 
      break;
    }

    retMap.insert("mode", modeStr);
#else
    retMap.insert("supported", false);
#endif
    return retMap;
}

/*!
    Sets "action" or "tab" as \a mode of the toolbar.
*/
void JSToolbar::setMode(QString mode)
{
#ifdef QTWRT_ENABLE_TABMODE_TOOLBAR
    QList<Action*> actions =
      m_actionAdapter->actions(Action::ToolBarLocation);

    if (mode.compare("action", Qt::CaseInsensitive) == 0) {
        foreach (Action* action, actions) {
            bool activated =
              action->isCheckable() && action->isChecked();
            action->setCheckable(false);
            // storing the state
            action->setActivated(activated);
        }
        m_mode = ToolbarAction;
    } else if (mode.compare("tab", Qt::CaseInsensitive) == 0) {
        foreach (Action* action, actions) {
            action->setCheckable(true);
            if (action->isActivated()) {
                action->setChecked(true);
            }
        }
        m_mode = ToolbarTab;
    } else {
        return;
    }
    
    emit toolbarModeChanged(m_mode);
#endif
}

/*!
    Appends given native instance \a id to menu.
*/
void JSToolbar::append(int id)
{
    if (m_actionAdapter) {
        JSMenuItem* item = JSMenuItemFactory::item(id);
        if (item) {
            Action* menuItem = item->action();
            menuItem->setLocation(Action::ToolBarLocation);
#ifdef QTWRT_ENABLE_TABMODE_TOOLBAR
            if (m_mode == ToolbarTab) {
                menuItem->setCheckable(true);
                if (menuItem->isActivated())
                    menuItem->setChecked(true);
            }
#endif
            m_actionAdapter->addAction(menuItem);
            m_mainLevelMenu.append(menuItem);            
        }
    }
}

/*!
    Removes given native instance \a id from menu.
*/
void JSToolbar::remove(int id)
{
    if (m_actionAdapter) {
        JSMenuItem* item = JSMenuItemFactory::item(id);
        if (item) {
            Action* menuItem = item->action();
#ifdef QTWRT_ENABLE_TABMODE_TOOLBAR
            if (m_mode == ToolbarTab) {
                bool activated = 
                  menuItem->isCheckable() && menuItem->isChecked();
                menuItem->setActivated(activated);
            }
            menuItem->setCheckable(false);
            menuItem->setChecked(false);
#endif
            m_actionAdapter->removeAction(menuItem);
            m_mainLevelMenu.removeAll(menuItem);
        }
    }
}

/*!
    Inserts given native instance \a id to menu \a beforeIndex.
*/
void JSToolbar::insert(int id, int beforeIndex) {
    if (m_actionAdapter) {
        JSMenuItem* item = JSMenuItemFactory::item(id);
        if (item) {
            Action* menuItem = item->action();
            menuItem->setLocation(Action::ToolBarLocation);
#ifdef QTWRT_ENABLE_TABMODE_TOOLBAR
            if (m_mode == ToolbarTab) {
                menuItem->setCheckable(true);
                if (menuItem->isActivated())
                    menuItem->setChecked(true);
            }
#endif
            
            // Append invalid indexes at the end, then menuItem is appended into menu.
            if (beforeIndex >= 0 && beforeIndex < m_mainLevelMenu.length()) {
                Action* beforeItem = m_mainLevelMenu.at(beforeIndex);
                m_actionAdapter->insertAction(beforeItem, menuItem);
            } else {
                m_actionAdapter->addAction(menuItem);
                m_mainLevelMenu.append(menuItem);
            }
        }
    }
}

/*!
    Clears menu items from menu.
*/
void JSToolbar::clear() {
    if (m_actionAdapter) {
        QList<Action*> actionsToClear = m_actionAdapter->actions(Action::ToolBarLocation);
#ifdef QTWRT_ENABLE_TABMODE_TOOLBAR
        foreach (Action* action, actionsToClear) {
            if (m_mode == ToolbarTab) {
                bool activated = 
                    action->isCheckable() && action->isChecked();
                action->setActivated(activated);
            }
            action->setCheckable(false);
            action->setChecked(false);
        }
#endif
        m_actionAdapter->removeActions(actionsToClear);
        m_mainLevelMenu.clear();
    }
}

}
