/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include <QDir>
#include <QMapIterator>

#include "wrtpage.h"
#include "qwebframe.h"
#include "serviceproxy.h"

#include "wrtapiinterface.h"
#include "jsmenuitemfactory.h"
#include "jsviewmenu.h"
#include "jscontextmenu.h"
#include "jswidgetinterface.h"
#include "jsresourcewrapper.h"
#include "jsversion.h"
#include "jstoolbar.h"
#include "jsescapebutton.h"
#include "jstheme.h"

#include "wrtapiprovider.h"

namespace WRT {


WrtApiProviderPrivate::WrtApiProviderPrivate(WrtApiProvider* wrtApiProvider)
    : q(wrtApiProvider)
{
}

WrtApiProviderPrivate::~WrtApiProviderPrivate()
{
    JSResourceWrapper* api;
    while (!deleteList.isEmpty()) {
        api = static_cast<JSResourceWrapper*>(deleteList.takeFirst());
        delete api;
    }
}

void WrtApiProviderPrivate::finalizeApi(JSResourceWrapper* resourceWrapper, JSApi::WidgetJavaScriptApiFlag api)
{
    QObject* apiObject = resourceWrapper;

    if (resourceWrapper) {
        resourceWrapper->setWebPage(q->getContext());
        if (!resourceWrapper->name().isEmpty()) {
            // TODO : QScriptEngine ownership must be added.
            q->addService(resourceWrapper->name(), *apiObject);
        }
        m_apis.insert(api, resourceWrapper);
        deleteList.append(apiObject);
    }

}

WrtApiProvider::WrtApiProvider()
    : d(new WrtApiProviderPrivate(this))
{
}


WrtApiProvider::~WrtApiProvider()
{
    destroyInstance();
    foreach (QString key, d->m_services.keys()) {
        ServiceProxy *proxy = d->m_services.value(key);
        d->m_services[key] = NULL;
        delete proxy;
    }

    delete d;
}


void WrtApiProvider::setApis(JSApi::WidgetJavaScriptApiFlags apis)
{
    d->m_apiNames = apis;
}


void WrtApiProvider::handleNewInstance()
{

    // Evaluate CJSE device.js as a first JavaScript file
    if (d->m_apiNames & JSApi::Device) {
        JSResourceWrapper * jsResource = new JSResourceWrapper("wrt/device/device.js", "");
        d->finalizeApi(jsResource, JSApi::Device);
    }

    // Evaluate common JavaScript utils before evaluating any other WRT API.
    JSResourceWrapper * jsResource = new JSResourceWrapper("utils.js", "");
    d->finalizeApi(jsResource, JSApi::Utils);

    if (d->m_apiNames & JSApi::MenuItem) {
        JSMenuItemFactory* menuItemFactory = new JSMenuItemFactory();
        d->finalizeApi(menuItemFactory, JSApi::MenuItem);
    }

    if (d->m_apiNames & JSApi::Menu) {
        JSViewMenu* menu = new JSViewMenu();
        d->finalizeApi(menu, JSApi::Menu);
    }

    if (d->m_apiNames & JSApi::ContextMenu) {
        JSContextMenu* contextMenu = new JSContextMenu();
        d->finalizeApi(contextMenu, JSApi::ContextMenu);
    }

    if (d->m_apiNames & JSApi::Toolbar) {
        JSToolbar* toolbar = new JSToolbar();
        d->finalizeApi(toolbar, JSApi::Toolbar);
    }

    if (d->m_apiNames & JSApi::EscapeButton) {
        JSEscapeButton * escapeButton = new JSEscapeButton();
        d->finalizeApi(escapeButton, JSApi::EscapeButton);
    }

    if (d->m_apiNames & JSApi::WidgetInterface) {
        JSWidgetInterface * widgetInterface = new JSWidgetInterface();
        d->finalizeApi(widgetInterface, JSApi::WidgetInterface);
    }

    if (d->m_apiNames & JSApi::Version) {
        JSVersion * version = new JSVersion();
        d->finalizeApi(version, JSApi::Version);
    }

    if (d->m_apiNames & JSApi::Theme) {
        JSTheme* theme = new JSTheme();
        d->finalizeApi(theme, JSApi::Theme);
    }
}


void WrtApiProvider::destroyInstance()
{
    // FIXME : Cleanup code here
}

void WrtApiProvider::pageOnReload()
{
    WrtApiInterface* apiInterface;
    // Order is defined by the key -> numbers at wrtglobal.h
    QMapIterator<JSApi::WidgetJavaScriptApiFlag, WrtApiInterface*> iterator(d->m_apis);
    while (iterator.hasNext()) {
        iterator.next();
        apiInterface = iterator.value();
        if (apiInterface) {
            apiInterface->pageOnReload();
        }
    }
}

WrtApiInterface* WrtApiProvider::apiObject(JSApi::WidgetJavaScriptApiFlag api)
{
    return d->m_apis.value(api, 0);
}

void WrtApiProvider::windowObjectCleared()
{
    if (d->m_page != NULL) {
        QMap<QString, ServiceProxy*>::iterator i = d->m_services.begin();
        while (i != d->m_services.end()) {
            d->m_page->mainFrame()->addToJavaScriptWindowObject(i.key(),i.value(), i.value()->objectOwnership());
            ++i;
        }
    }

    pageOnReload();
}

bool WrtApiProvider::addService(const QString& serviceName, QObject& newObj, bool giveToGarbageCollector)
{
    // save the service objects
    if (d->m_services.contains(serviceName))
        return FALSE;
    // create service proxy so you can marshall
    // the calls

    ServiceProxy *proxy = new ServiceProxy(newObj, giveToGarbageCollector);
    proxy->initialize();
    d->m_services.insert(serviceName, proxy);
    return true;

}

void WrtApiProvider::setContext(WrtPage* page)
{
    d->m_page = page;
    connect(page->mainFrame(),
            SIGNAL(javaScriptWindowObjectCleared()),
            this, 
            SLOT(windowObjectCleared()));
}

WrtPage* WrtApiProvider::getContext()
{
    return d->m_page;
}

}

