/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <QWidget>
#include <QMessageBox>
#include <QDir>
#include <QtAlgorithms>

#include "WidgetManager_p.h"
#include "widgetmanagerconstants.h"
#include "proprietarytags.h"
#include "WidgetInfo.h"
#include "W3CXmlParser/w3cxmlplugin.h"
#include "private/WidgetUtilsLogs.h"

#include "webappinfo.h"
#include "webappuniquenesshelper.h"
#include "WebAppRegistry.h"
#include "WidgetProperties.h"
#include "WidgetManager.h"

#include "private/webappmanager_p.h"

#include "signatureparser.h"
#include "SuperWidget.h"

#include "WgtWidget.h"
#include "widgetpromptstrings.h"
#include <QDesktopServices>
#include <QThreadPool>

#define NO_NAME "NoName"


/*!
 \class WidgetManager
 \brief  Install and uninstall web applications. Public class for installing widgets for 3rd party.
*/

/*!
Widget manager deletes all temporary created files on destructor

\a widget pointer for displaying confirmation
*/
Q_DECL_EXPORT WidgetManager::WidgetManager(QWidget *parent, bool isChromeless)
    : QObject(parent)
    , d(new WidgetManagerPrivate(parent, this))
{
    (void)isChromeless; // Argument required for compatability with old branch
    d->m_widgetType = WidgetTypeUnknown;
    d->m_contentType = CONTENT_TYPE_UNKNOWN;
    d->m_installingWidget = 0;
}


Q_DECL_EXPORT WidgetManager::~WidgetManager()
{
    delete d->m_installingWidget;

    qDeleteAll(d->m_widgetCache.values());

#if defined(QTWRT_USE_USIF)
    if (d->iAppInfo)
        delete d->iAppInfo;
#endif

    delete d;
}

/*!
 already installed web applications executable file/path

 \a pkgPath package file path
 \return returns launcher file path (html file path for widgets) or link for web applications
*/
Q_DECL_EXPORT QString WidgetManager::launcherPath(const QString &pkgPath)
{
    if (!d->checkCallerPermissions()) {
        return "";
    }

    LOG("BEGIN WidgetManager::launcherPath(" << pkgPath << ")");
    // Get the widget
    SuperWidget* widget = getWidgetFromCache(pkgPath);
    if (!widget) {
        // We don't have a widget for this path yet, create one
        widget = d->createWidget(pkgPath);
    }
    if (widget) {
        LOG("END WidgetManager::launcherpath returns" << widget->launcherPath(pkgPath));
        return widget->launcherPath(pkgPath);
    }

    LOG("END WidgetManager::launcherPath returns ' '");
    return "";
}

/*!
 \a pkgPath package file path
 \return returns true if pkgPath is already installed
*/
Q_DECL_EXPORT bool WidgetManager::isRegistered(const QString &pkgPath)
{
    if (!d->checkCallerPermissions()) {
        return false;
    }

    WidgetProperties *widgetProp = d->getWidgetProperties(pkgPath);
    if (widgetProp) {
        LOG("WidgetManager::isRegistered : pkgPath: " << pkgPath << WebAppRegistry::instance()->isRegistered (widgetProp->id()));
        return WebAppRegistry::instance()->isRegistered (widgetProp->id());
    }

    return false;
}

/*!
 install's the package file passed. Registers the application in registry if package file is valid
o
 \a pkgPath package file path
 \return returns true if installation is successful
*/
WidgetInstallError WidgetManager::install(const QString& pkgPath,
                                          QString& widgetId,
                                          bool silent,
                                          bool update,
                                          const QString& rootDirectory)
{
    if (!d->checkCallerPermissions()) {
        return WidgetInstallPermissionFailed;
    }

    LOG("WidgetManger::install silent =" << silent << "update =" << update);

    if (!d->validateDriveLetter(rootDirectory)) {
        return WidgetDriveLetterValidationFailed;
    }
    if (update) {
        return install(pkgPath, widgetId, "", d->m_rootDirectory, true, true);
    }
    return install(pkgPath, widgetId, "", d->m_rootDirectory, silent, false);
}

void WidgetManager::tryContinueInstallation() {
    if (d->m_installingWidget) {
        d->m_installingWidget->tryContinueInstallation();
    }
}

Q_DECL_EXPORT WidgetProperties* WidgetManager::getProperties(const QString &pkgPath)
{
    if (!d->checkCallerPermissions()) {
        return 0;
    }

    return d->getWidgetProperties(pkgPath);
}

SuperWidget* WidgetManager::getWidgetFromCache(QString pkgPath)
{
    LOG("BEGIN WidgetManager::getWidgetFromCache");
    QString uniqueName;
    if (WebAppUniquenessHelper::getUniqueNameFromPath(pkgPath, uniqueName)) {
        LOG("WidgetManager::getWidgetFromCache uniqueName =" << uniqueName);

        if (d->m_widgetCache.contains(uniqueName)) {
            SuperWidget* widget = d->m_widgetCache.value(uniqueName);
            LOG("END WidgetManager::getWidgetFromCache found a widget =" << widget);
            return widget;
        }
    }

    LOG("END WidgetManager::getWidgetFromCache can't find widget");
    return 0;
}

void WidgetManager::setWidgetIntoCache(QString pkgPath, SuperWidget* widget)
{
    LOG("BEGIN WidgetManager::setWidgetIntoCache");
    QString uniqueName;
    if (WebAppUniquenessHelper::getUniqueNameFromPath(pkgPath, uniqueName)) {
        LOG("WidgetManager::setWidgetIntoCache uid =" << uniqueName);

        if (!d->m_widgetCache.contains(uniqueName)) {
            d->m_widgetCache.insert(uniqueName, widget);
            LOG("WidgetManager::setWidgetIntoCache saved widget =" << widget);
        }
    }
    LOG("END WidgetManager::setWidgetIntoCache");
}

void WidgetManager::removeWidgetFromCache(QString pkgPath)
{
    QString uniqueName;
    if (WebAppUniquenessHelper::getUniqueNameFromPath(pkgPath, uniqueName)) {

        if (d->m_widgetCache.contains(uniqueName)) {
            SuperWidget* widget = d->m_widgetCache.value(uniqueName);
            delete widget;
            d->m_widgetCache.remove(uniqueName);
        }
    }
}

void WidgetManager::setInstallationAttributes(WRT::InstallationAttributes attributes) {
    d->m_attributes = attributes;
}

// function to install widgets
// parameters:
//     widgetPath          path to widget bundle
//     widgetContentType   content type of widget (optional)
//     silent              true for silent install (optional)
//                         false by default
// return
//     bool                true if successful
//
// creates appropriate widget and installs it
WidgetInstallError WidgetManager::install(const QString& widgetPath,
                                          QString &widgetId,
                                          const QString& widgetContentType,
                                          const QString& rootDirectory,
                                          const bool silent,
                                          const bool update,
                                          const QString &sigId)
{
    Q_UNUSED(sigId);
    d->m_silent = silent;

    WidgetType widgetType = SuperWidget::getWidgetType(widgetPath, widgetContentType);

    if (WidgetTypeUnknown == widgetType) {
        LOG("WidgetManager::install - Invalid Widget type");
        d->handleInstallError(WidgetTypeValidationFailed);
        emit installationError(WidgetTypeValidationFailed);
        return WidgetTypeValidationFailed;
    }

    if (d->m_installingWidget) {
        delete d->m_installingWidget;
        d->m_installingWidget = 0;
    }

    // Create a new installing widget
    QString rootDir = rootDirectory;

    switch (widgetType) {
    case WidgetTypeW3c:
        d->m_installingWidget = new WgtWidget(rootDir);
        d->m_installingWidget->disableBackupRestoreValidation(d->m_disableUnsignWidgetSignCheck);
        d->m_installingWidget->setInstallationAttributes(d->m_attributes);
        break;
    default:
        LOG("WidgetManager::install Invalid Widget type");
        return WidgetTypeValidationFailed;
    }

    LOG("WidgetManager::install widget installed path: " << d->m_installingWidget->widgetInstallPath());

    if (!d->m_installingWidget) {
        return WidgetSystemError;
    }

    // FIXME : Silent / not silent should not visible. SIGNALS or interfacing for this class
    // Symbian OS installs can prompt after in-memory config.xml parsing in order
    // to be able to show widget name, etc., rather than filename
    if (!silent) {
        bool cancelled = !d->showQuestion(TR_WM_WIDGET_INSTALL,
                                          TR_WM_WIDGET_INSTALL_QUERY + QFileInfo(widgetPath).fileName() + " ?",
                                          d->WidgetManagerPromptButtonYes,
                                          d->WidgetManagerPromptButtonNo);

        if (cancelled)
        {
            d->showInformation(TR_WM_WIDGET_INSTALL, TR_WM_WIDGET_INSTALL_CANCELLED);

            return WidgetUserConfirmFailed;
        }
    }

    connect(d->m_installingWidget, SIGNAL(installProgress(int)), this, SIGNAL(installProgress(int)));
    connect(d->m_installingWidget,
            SIGNAL(installProgress(int)),
            d,
            SLOT(handleInstallProgress(int)),
            Qt::DirectConnection);

    // Connect all signals from a widget to dialgos handles of this class.
    connect(d->m_installingWidget, SIGNAL(installationError(WidgetInstallError)),
            d, SLOT(handleInstallError(WidgetInstallError)));
    connect(d->m_installingWidget, SIGNAL(aboutToReplaceExistingWidget(QString)),
            d, SLOT(handleWidgetReplacement(QString)), Qt::DirectConnection);
    connect(d->m_installingWidget, SIGNAL(queryInstallationDestination(unsigned long, bool)),
            d, SLOT(handleDestinationSelection(unsigned long, bool)), Qt::DirectConnection);
    connect(d->m_installingWidget, SIGNAL(aboutToInstallWidgetWithFeatures(QList<QString>)),
            d, SLOT(handleFeatureInstallation(QList<QString>)),Qt::DirectConnection);
    connect(d->m_installingWidget, SIGNAL(aboutToInstallUntrustedWidget(WidgetInstallError)),
            d, SLOT(handleUntrustedWidgetInstallation(WidgetInstallError)), Qt::DirectConnection);
    connect(d->m_installingWidget, SIGNAL(installationSucceed()),
            d, SLOT(handleInstallationSuccess()));

    // Widget installation signals forwarded out of this class.
    // This provides one way to rid of qmessageboxes from this class and applications that are
    // using this would be in charge of showing correct ui dialogs.
    // However, silentInstall is only thingie which could really utilize this.
    connect(d->m_installingWidget, SIGNAL(installationError(WidgetInstallError)),
            this, SIGNAL(installationError(WidgetInstallError)));
    connect(d->m_installingWidget, SIGNAL(aboutToReplaceExistingWidget(QString)),
            this, SIGNAL(aboutToReplaceExistingWidget(QString)), Qt::DirectConnection);
    connect(d->m_installingWidget, SIGNAL(queryInstallationDestination(unsigned long, bool)),
            this, SIGNAL(queryInstallationDestination(unsigned long, bool)), Qt::DirectConnection);
    connect(d->m_installingWidget, SIGNAL(aboutToInstallUntrustedWidget(WidgetInstallError)),
            this, SIGNAL(aboutToInstallUntrustedWidget(WidgetInstallError)), Qt::DirectConnection);
    connect(d->m_installingWidget, SIGNAL(aboutToInstallWidgetWithFeatures(QList<QString>)),
            this, SIGNAL(aboutToInstallWidgetWithFeatures(QList<QString>)),Qt::DirectConnection);
    connect(d->m_installingWidget, SIGNAL(installationSucceed()),
            this, SIGNAL(installationSucceed()));

    enum WidgetInstallError status=WidgetInstallSuccess;
    d->m_installingWidget->setWidgetBundlePath(widgetPath);
    if (update) {
        status = d->m_installingWidget->install(true);
        disconnect(d->m_installingWidget,
                   SIGNAL(installProgress(int)),
                   this,
                   SIGNAL(installProgress(int)));
        if (status == WidgetInstallSuccess) {
            widgetId = d->m_installingWidget->getProperties()->id();
            removeWidgetFromCache(d->m_installingWidget->getProperties()->installPath());
        }

        return status;
    }

    if (!sigId.isNull() && !sigId.isEmpty() && widgetType == WidgetTypeW3c) {
        status = d->m_installingWidget->install();
    } else {
        status = d->m_installingWidget->install();
    }

    if (status == WidgetInstallSuccess) {
        widgetId = d->m_installingWidget->getProperties()->id();
    }

    disconnect(d->m_installingWidget,
               SIGNAL(installProgress(int)),
               this,
               SIGNAL(installProgress(int)));

    return status;
}

/*
 * Returns the installing widget properties
 */
WidgetProperties* WidgetManager::getProperties()
{
    if (!d->checkCallerPermissions()) {
        return 0;
    }

    if (d->m_installingWidget)
    {
        return d->m_installingWidget->getProperties();
    }
    return 0;
}

/*
 * Constructs the widgets installation path
 * TODO: This method actually gets the data location, not installation path.
 *       Should be renamed getWidgetDataPath().
 */
QString WidgetManager::getWidgetInstallPath(const QString& rootDirectory)
{
    Q_UNUSED(rootDirectory);
    QString installPath("");

#if defined(Q_OS_MAEMO5) || defined(Q_OS_MAEMO6)
    installPath = WIDGET_INSTALL_PATH;
#endif

    if (! createDir(installPath)) {
        LOG("SuperWidget::getWidgetInstallPath - Could not create widget dir:" << installPath);
        return QString();
    }
  return installPath;
}

/*
 * Directory Creation for Widget Installation
 */
bool WidgetManager::createDir(const QString& path)
{
  QDir dir(path);
  if (!dir.exists()){
    if (!dir.mkpath(dir.absolutePath())) {
      LOG("WidgetManager::createDir could not create dir");
      return false;
      }
    }
    return true;
}

void WidgetManager::uninstallProgress(int progress)
{
}
