/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <QWidget>
#include <QMessageBox>
#include <QDir>
#include <QtAlgorithms>

#include "WidgetManager_p.h"
#include "widgetmanagerconstants.h"
#include "proprietarytags.h"
#include "WidgetInfo.h"
#include "W3CXmlParser/w3cxmlplugin.h"
#include "private/WidgetUtilsLogs.h"

#include "webappinfo.h"
#include "webappuniquenesshelper.h"
#include "WebAppRegistry.h"
#include "WidgetProperties.h"
#include "WidgetManager.h"

#include "private/webappmanager_p.h"

#define PDATA ((WidgetManager_p*) m_data)

#include "signatureparser.h"
#include "SuperWidget.h"

#include "WgtWidget.h"
#include "widgetpromptstrings.h"
#include <QDesktopServices>
#include <QThreadPool>

#define NO_NAME "NoName"

WidgetManager_p::WidgetManager_p(QWidget *parent):
    m_parent(parent)
{
}

WidgetManager_p::~WidgetManager_p()
{
}

WidgetManagerPrivate::WidgetManagerPrivate(QWidget *parent, WidgetManager* widgetManager)
    : m_silent(false)
    , m_disableUnsignWidgetSignCheck(false)
    , m_attributes(WRT::NoAttibutes)
    , m_parent(parent)
    , q(widgetManager)
{
    m_data = (void *) new WidgetManager_p(parent);
}

WidgetManagerPrivate::~WidgetManagerPrivate()
{
    delete (WidgetManager_p*)m_data;
}

//
// function to get widget properties from path. Will create a new widget object
// if one doesn't exist for the path
// parameters:
//     path       path to widget
// return
//     WidgetProperties* pointer to widget properteis
//
WidgetProperties* WidgetManagerPrivate::getWidgetProperties(const QString &pkgPath)
{
    LOG("BEGIN WidgetManagerPrivate::getWidgetProperties pkgpath =" << pkgPath);

    if (pkgPath.isEmpty()) {
        LOG("END WidgetManagerPrivate::getWidgetProperties pkgPath is empty");
        return 0;
    }

    SuperWidget* cachedWidget = q->getWidgetFromCache(pkgPath);
    if (cachedWidget) {
       LOG( "WidgetManagerPrivate::getWidgetProperties use cached widget to getProperties");
        WidgetProperties *props = cachedWidget->getProperties(pkgPath);
        if (props == 0) {
            LOG("END WidgetManagerPrivate::getWidgetProperties: cached widget has properties=0");
            return 0;
        }
        // Check if current widget path and properties path match. Use native
        // separators and remove trailing separator, because properties path
        // was saved using native separators.
        QString widgetPath = QDir::convertSeparators(pkgPath);

        QStringList widgetPathParts = widgetPath.split(QDir::separator(),QString::SkipEmptyParts);
        QString realWidgetPath = widgetPathParts.join(QDir::separator());
        if (widgetPath.startsWith(QDir::separator())) {
            realWidgetPath = QDir::separator() + realWidgetPath;
        }

        QStringList installPathParts = props->installPath().split(QDir::separator(),QString::SkipEmptyParts);
        QString realInstallPath = installPathParts.join(QDir::separator());
        if (props->installPath().startsWith(QDir::separator())) {
            realInstallPath = QDir::separator() + realInstallPath;
        }
        if (realInstallPath.compare(realWidgetPath) == 0) {
            LOG("END WidgetManagerPrivate::getWidgetProperties cached widget properties are valid");
            return props;
        }
    }
    else {
        // We couldn't find a cached widget, so create a new one
        SuperWidget* newWidget = createWidget(pkgPath);
        if (newWidget) {
            WidgetProperties *props = newWidget->getProperties(pkgPath);
            LOG("END WidgetManagerPrivate::getWidgetProperties using new widget to get properties");
            return props;
        }
    }

    LOG("END WidgetManagerPrivate::getWidgetProperties FAILED to get widget properties");
    return 0;

}

//
// function to create a widget, uses the path (unique folder name) as key to
// cache widget in hash
// parameters:
//     path       path to widget
// return
//     SuperWidget* pointer to new widget
//
SuperWidget* WidgetManagerPrivate::createWidget(const QString& pkgPath)
{
    // Generate the install path, use native separators and no ending separator
    QString installPath;
#if defined(Q_OS_LINUX)
    installPath = WIDGET_INSTALL_PATH;
#endif

    // Create a new widget
    SuperWidget* newWidget;
    switch (SuperWidget::getWidgetType(pkgPath))
        {
        case WidgetTypeW3c:
            newWidget = new WgtWidget(installPath);
            break;
        default:
            LOG("END WidgetManagerPrivate::createWidget - Invalid Widget type");
            newWidget = 0;
            break;
    }

    if (newWidget) {
        q->setWidgetIntoCache(pkgPath, newWidget);
        return newWidget;
    }

    return 0;
}

void WidgetManagerPrivate::handleInstallProgress(int p) {
  qDebug() << "Install progress: " << p;
}

void WidgetManagerPrivate::handleInstallError(WidgetInstallError error) {
    // FIXME : We must localize message and use platform specific dialogs / notifications.
    if (!m_silent && ((error != WidgetInstallSuccess) && (error!= WidgetUserConfirmFailed))) {
        showInformation(TR_WM_WIDGET_INSTALL_ERROR, TR_WM_WIDGET_INSTALL_FAILED);
    }
}

void WidgetManagerPrivate::handleInstallationSuccess() {
}

void WidgetManagerPrivate::handleWidgetReplacement(QString title) {
    if (!m_silent) {
        QString msg = TR_WM_WIDGET_REPLACE + " \" " + title + " \"?";
        if (!showQuestion(TR_WM_WIDGET_INSTALL,
                          msg,
                          WidgetManagerPromptButtonYes,
                          WidgetManagerPromptButtonNo))
        {
            return;
        }
    }
    m_installingWidget->tryContinueInstallation();
}

void WidgetManagerPrivate::handleFeatureInstallation(QList<QString> capList)
{
    if (!m_silent) {
        QString message;
        for (int i=0; i<capList.count(); i++) {
            message += "\n";
            message += QCoreApplication::translate("WidgetManager", QT_TRANSLATE_NOOP("WidgetManager",capList[i].toUtf8().constData()));
        }

        if (!showQuestion(TR_WM_SECURITY_WARNING,
                             TR_WM_WIDGET_INSTALL_ALLOW +
                             m_installingWidget->getProperties()->title() + TR_WM_WIDGET_INSTALL_TO_ACCESS + message,
                             WidgetManagerPromptButtonOkay,
                             WidgetManagerPromptButtonCancel))
        {
            return;
        }
    }
    m_installingWidget->tryContinueInstallation();
}

void WidgetManagerPrivate::handleDestinationSelection(unsigned long spaceRequired, bool allowRemovable) {
    if (m_silent) {
        m_installingWidget->tryContinueInstallation();
        return;
    }

    m_installingWidget->tryContinueInstallation();
}

void WidgetManagerPrivate::handleUntrustedWidgetInstallation(WidgetInstallError errorCode) {

    if (m_silent) {
        // Same code as use to be in the WgtWidget installation.
        // In silent mode must enable installation of untrusted widget.
        // This is just about the error value.
        if (errorCode == WidgetFindSignatureFailed) {
            m_installingWidget->tryContinueInstallation();
        }
        return;
    }

    QString message = "";
    switch (errorCode) {
        case WidgetFindSignatureFailed:
        case WidgetSignatureParsingFailed:
        case WidgetSignatureOrSignedInfoMissing:
        case WidgetSignatureRefExistFailed:
        case WidgetCertValidationFailed:
        case WidgetSignatureValidationFailed:
        case WidgetSignatureRefValidationFailed:
            message = TR_WM_WIDGET_SIGNATURE_INVALID;
            break;
        default:
            message = TR_WM_WIDGET_UNKNOWN_ERROR;
            break;
    }

    if (showQuestion(TR_WM_WIDGET_INSTALL,
                     message + TR_WM_WIDGET_INSTALL_CONT_QUERY,
                     WidgetManagerPromptButtonYes,
                     WidgetManagerPromptButtonNo))
    {
            m_installingWidget->tryContinueInstallation();
    }
}

/*
 * Checks for the rootDirectory value
 * If Null or Empty then it assigns the DEFAULT_ROOT_DIRECTORY
 */
QString WidgetManagerPrivate::setDefaultRootDirectory(const QString& rootDirectory)
{
    if (rootDirectory.isEmpty() || rootDirectory.isNull()){
        QString defaultRootDir = DEFAULT_ROOT_DIRECTORY;
        return(defaultRootDir);
    }
    QString& defaultRootDir = const_cast<QString&> (rootDirectory);
    return(defaultRootDir);
}

//
// function to validate the input rootDirectory string
// parameters:
//     QString    rootDirectory String
// return
//     bool    True if input is valid on Symbian
//
bool WidgetManagerPrivate::validateDriveLetter(const QString& rootDirectory)
{
    m_rootDirectory = setDefaultRootDirectory(rootDirectory);
    m_rootDirectory = q->getWidgetInstallPath(m_rootDirectory);
    if (m_rootDirectory.isEmpty() || m_rootDirectory.isNull())
        return false;

    return true;
}

bool WidgetManagerPrivate::checkCallerPermissions()
{
    return true;
}

void WidgetManagerPrivate::showInformation(QString title, QString msg)
{
    QMessageBox::information(PDATA->m_parent, title, msg);
}


bool WidgetManagerPrivate::showQuestion(QString title,
                                        QString msg,
                                        WidgetManagerPromptButtons primaryButton,
                                        WidgetManagerPromptButtons secondaryButton)
{
    QMessageBox::StandardButtons buttons = QMessageBox::NoButton;
    QMessageBox::StandardButton primary = QMessageBox::NoButton;

    switch (primaryButton) {
    case WidgetManagerPromptButtonYes:
        primary = QMessageBox::Yes;
        break;
    case WidgetManagerPromptButtonNo:
        primary = QMessageBox::No;
        break;
    case WidgetManagerPromptButtonCancel:
        primary = QMessageBox::Cancel;
        break;
    case WidgetManagerPromptButtonOkay:
    default:
        primary = QMessageBox::Ok;
        break;
    }

    buttons |= primary;

    switch (secondaryButton) {
    case WidgetManagerPromptButtonYes:
        buttons |= QMessageBox::Yes;
        break;
    case WidgetManagerPromptButtonNo:
        buttons |= QMessageBox::No;
        break;
    case WidgetManagerPromptButtonOkay:
        buttons |= QMessageBox::Ok;
        break;
    case WidgetManagerPromptButtonCancel:
    default:
        buttons |= QMessageBox::Cancel;
        break;
    }

    return (primary == QMessageBox::question(PDATA->m_parent, title, msg, buttons));
}

void WidgetManagerPrivate::uninstallComplete(WidgetUninstallError errorCode)
{
}
