/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "debianutils.h"
#include "desktopfilewriter.h"
#include "SuperWidget.h"
#include "archiver.h"
#include "unzip.h"
#include "w3csettingskeys.h"
#include "wbenchmark.h"
#include "wrtsettings.h"
#include "zzip.h"

#include <QDateTime>
#include <QDesktopServices>
#include <QProcess>
#include <QStringList>
#include <QXmlStreamWriter>
#include <sys/stat.h>


DebianUtils::DebianUtils(SuperWidget* widget, const QString packageName, const QString& sourcePath, const QString& installationPath, const QString& appId)
    : PackageUtils(widget, packageName, sourcePath, installationPath, appId)
{
    m_packageFileName = m_packageName + ".deb";
    m_packageFilePath = m_packageRootPath + QDir::separator() + m_packageFileName;
}

bool DebianUtils::createPackage()
{
    if (!preCreatePackage() || !writeControlFile())
        return false;

    WBM("creating Debian package");
    qDebug() << "Creating deb file" << m_packageName << ".deb";
    bool bRet = true;

    // Create control file
    WBM_BEGIN("creating control.tar.gz", control);
    Archiver control(m_packageRootPath + "/control.tar.gz", Archiver::TarGz);
    bRet = bRet && control.isValid();
    bRet = bRet && control.addFileFromBuffer(m_controlFile, "./control", 444);
#if defined(Q_OS_MAEMO5)
    if (!m_preinstScript.isEmpty())
        bRet = bRet && control.addFileFromBuffer(m_preinstScript, "./preinst", 444);
#endif
    bRet = bRet && control.addFileFromBuffer(m_postinstScript, "./postinst", 444);
    bRet = bRet && control.addFileFromBuffer(m_postrmScript, "./postrm", 444);
    bRet = bRet && control.flush();
    if (!bRet) {
        qCritical() << "FAIL: creating control.tar.gz";
        return false;
    }
    WBM_END(control);

    // Create data file
    WBM_BEGIN("creating data.tar", data);
    Archiver data(m_packageRootPath + "/data.tar", Archiver::Tar);
    bRet = data.isValid();
    if (bRet) {
        QStringList files = QDir(m_packageRootPath).entryList(QDir::Dirs | QDir::NoDotAndDotDot);
        foreach (QString file, files) {
#if defined(Q_OS_MAEMO5)
            if (file == "unzip")
                continue;
#endif
            if (bRet)
                bRet = data.addDirectoryRecursively(m_packageRootPath + "/" + file, file);
        }
        // add some files here
        bRet = bRet && data.flush();
        if (!bRet) {
            qCritical() << "FAIL: creating data.tar";
            return false;
        }
    }
    WBM_END(data);

    // Create debian file
    WBM_BEGIN("creating ar archive", ar);
    Archiver debian(m_packageRootPath + QDir::separator() + m_packageName + ".deb", Archiver::Ar);
    bRet = data.isValid();
    if (bRet) {
        bRet = bRet && debian.addFileFromBuffer("2.0\n", "debian-binary", 444);
        bRet = bRet && debian.addFile(m_packageRootPath + "/control.tar.gz", "control.tar.gz");
        bRet = bRet && debian.addFile(m_packageRootPath + "/data.tar", "data.tar");
        bRet = bRet && debian.flush();
        if (!bRet) {
            qCritical() << "FAIL: creating debian file";
            return false;
        }
    }
    WBM_END(ar);

    return true;
}

bool DebianUtils::writeControlFile()
{
    QDir dir;
    if (!dir.mkpath(m_packageRootPath))
        return false;

    DebianPackageMetadata meta;
    meta.packageName = m_packageName;
    meta.appDescription = m_appDescription;
    meta.appName = m_appName;
    meta.appVersion = m_appVersion;
    meta.authorName = m_authorName;
    meta.installedSize = m_installedSize;

    m_controlFile.append(generateControlFile( meta));

     return true;
}

QString DebianUtils::generateValidDebianVersion(const QString& originalVersion)
{
    QString validVersion = originalVersion;
    validVersion.remove(" ");
    if (!isValidDebianVersion(validVersion)) {
        // If supplied originalVersion string would fail in debian packaging,
        // we replace it with default version string + original version where illegal chars have been removed.
        validVersion = convertToAlphaNumeric(originalVersion);
    }

    return validVersion;
}

QString DebianUtils::convertToAlphaNumeric(const QString& original)
{
    QString converted = original;
    for (int i = 0; i < converted.size(); i++) {
        QChar c = converted.at(i);
        if (!c.isLetterOrNumber())
            converted.replace(i, 1, "_");
    }

    return converted;
}

bool DebianUtils::isValidDebianVersion(const QString& version)
{
    // Version must start with digit
    if (version.size() > 0 && !version.at(0).isDigit())
        return false;

    // If there is hyphen in the version string, debian_revision must exist
    if (version.contains("-") && !hasValidDebianRevision(version))
        return false;

    // If there is colon in the version string, epoch must exist
    if (version.contains(":") && !hasValidDebianEpoch(version))
        return false;

    // Check upstream_veersion.
    if (!hasValidDebianUpstreamVersion(version))
        return false;

    return true;
}

bool DebianUtils::hasValidDebianEpoch(const QString& version)
{
    bool valid = false;

    // If there is hyphen in the version string, debian_revision must exist
    int separatorIndex = version.indexOf(":");
    if ((separatorIndex > 0) && (separatorIndex < version.size())) {
        QString epoch = version.left(separatorIndex);
        epoch.toInt(&valid);
    }

    return valid;
}

bool DebianUtils::hasValidDebianUpstreamVersion(const QString& version)
{
    for (int i = 0; i < version.size(); i++) {
        QChar c = version.at(i);
        if (!c.isLetterOrNumber() && (c != '.') && (c != '+') && (c != '-') && (c != ':') && (c != '~'))
            return false;
    }
    return true;
}

bool DebianUtils::hasValidDebianRevision(const QString& version)
{
    bool valid = false;

    // If there is hyphen in the version string, debian_revision must exist
    int separatorIndex = version.lastIndexOf("-");
    if ((separatorIndex > 0) && (separatorIndex < version.size())) {
        QString revision = version.mid(separatorIndex);
        revision.toInt(&valid);
    }

    return valid;
}

QString DebianUtils::generateControlFile(const DebianPackageMetadata& packageMetadata)
{
    QString controlFile ("Package: " + packageMetadata.packageName + "\n"
                         "Source: " + packageMetadata.packageName + "\n"
                         "Priority: optional\n"
                         "Section: user/desktop\n"
                         "Maintainer: " + packageMetadata.authorName + "\n"
                         "Version: " + generateValidDebianVersion(packageMetadata.appVersion) + "\n"
                         "Architecture: all\n"
                         "Installed-Size: " + QString::number(packageMetadata.installedSize >> 10) + "\n"
                         "Maemo-Display-Name: " + packageMetadata.appName + "\n"
                         "Depends: qtwrt\n"
    // TODO: This is likely not working with more that 70 characters.
                         "Description: " + packageMetadata.appName + " (web application)" + "\n " + packageMetadata.appDescription.simplified() + "\n");
    return controlFile;

}
