/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "packageutils.h"
#include "desktopfilewriter.h"
#include "SuperWidget.h"
#include "widgetmanagerconstants.h"
#include "w3csettingskeys.h"
#include "wrtsettings.h"
#include "zzip.h"

#include <QDesktopServices>
#include <QFile>


PackageUtils::PackageUtils(SuperWidget* widget, const QString packageName, const QString& sourcePath, const QString &installationPath, const QString &appId)
    : m_widget(widget), m_packageName(packageName), m_sourcePath(sourcePath), m_installationPath(installationPath), m_appId(appId)
{
    m_packageRootPath = m_sourcePath.left(m_sourcePath.lastIndexOf(QDir::separator()));
    m_installedSize = 0;
}

PackageUtils::~PackageUtils()
{
    // do nothing
}

QString PackageUtils::packageFilePath() const
{
    return m_packageFilePath;
}

QString PackageUtils::packageFileName() const
{
    return m_packageFileName;
}

bool PackageUtils::preCreatePackage()
{
    bool result = setAttributes() && prepareWidget() && writeDesktopFile() && writeInstallationScripts() && writeRemovalScripts();

    return result;
}

bool PackageUtils::setAttributes()
{
    m_properties = m_widget->getProperties();

    // name
    m_appName = m_widget->value(W3CSettingsKey::WIDGET_NAME);
    if (m_appName.isEmpty())
        m_appName = QFileInfo(m_widget->widgetBundlePath()).completeBaseName();

    // version
    m_appVersion = m_widget->value(W3CSettingsKey::WIDGET_VERSION);
    if (m_appVersion.isEmpty())
       m_appVersion = DEFAULT_VERSION;

    // description
    m_appDescription = m_widget->value(W3CSettingsKey::WIDGET_DESCRIPTION);
    if (m_appDescription.isEmpty())
        m_appDescription = DEFAULT_DESCRIPTION;

    // author
    m_authorEmail = m_widget->value(W3CSettingsKey::WIDGET_AUTHOR, "email");
    m_authorName = m_widget->value(W3CSettingsKey::WIDGET_AUTHOR);

    // license
#if defined(Q_OS_MAEMO5)
    m_license = m_widget->value(W3CSettingsKey::WIDGET_LICENSE);
#endif

    // view modes
    QStringList viewModes = m_widget->value(W3CSettingsKey::WIDGET_VIEWMODES).split(" ");
    viewModes.removeDuplicates();

    foreach (QString viewMode, viewModes) {
        if (!SUPPORTED_HOMESCREEN_VIEW_MODES.contains(viewMode) && !SUPPORTED_APPLICATION_VIEW_MODES.contains(viewMode))
            viewModes.removeOne(viewMode);
    }
    if (viewModes.isEmpty())
        viewModes << DEFAULT_VIEW_MODE;

    foreach (QString viewMode, viewModes) {
        if (SUPPORTED_HOMESCREEN_VIEW_MODES.contains(viewMode)) {
            m_homescreenViewModes << viewMode;
        } else {
            m_applicationViewModes << viewMode;
        }
    }

    // icon
    m_iconPath = m_properties->iconPath();

    if (m_iconPath.right(4) == ".png" && !m_iconPath.endsWith(":/resource/default_widget_icon.png")) {
        m_iconName = m_appId;
#if !defined(Q_OS_MAEMO6)
        m_iconSymbolicLinkPath = ICON_SYMBOLIC_LINK_DIR + m_iconName + ".png";
#endif
    } else {
        m_iconName = DEFAULT_ICON;
    }

    return true;
}

bool PackageUtils::prepareWidget()
{
    QString widgetInstallPath(m_packageRootPath + m_installationPath);
    while (widgetInstallPath.endsWith(QDir::separator()))
        widgetInstallPath.chop(1);

    QDir dir;
    if (!dir.mkpath(widgetInstallPath.left(widgetInstallPath.lastIndexOf(QDir::separator()))))
        return false;

#if defined(Q_OS_MAEMO5)
    // dpkg on M5 doesn't support long path name, thus unzip the .wgt in postinst script
    m_installedSize += QFileInfo(m_widget->widgetBundlePath()).size();
    if (!dir.mkpath(widgetInstallPath)
        || !QFile::copy(m_widget->widgetBundlePath(), widgetInstallPath + QDir::separator() + QFileInfo(m_widget->widgetBundlePath()).fileName()))
        return false;
#else
    // move the unzipped widget to the temp folder
    m_installedSize += Zzip::uncompressedSize(m_widget->widgetBundlePath());
    if (!dir.rename(m_sourcePath, widgetInstallPath + QDir::separator()))
        return false;

    QDir baseRootDir(m_packageRootPath);
    baseRootDir.mkpath("etc/secure/s/" + m_packageName);
    baseRootDir.mkpath(m_properties->resourcePath());
#endif

    return true;
}

bool PackageUtils::writeDesktopFile()
{
    if (!m_properties->isSharedLibrary() || m_properties->isSharedLibraryWidget()) {
        DesktopFileWriter desktopFileWriter(m_packageRootPath, m_installationPath, m_appId);
        desktopFileWriter.setAppName(m_appName);
        desktopFileWriter.setAppIcon(m_iconName);
        desktopFileWriter.setHidden(m_properties->isHiddenWidget());
        desktopFileWriter.setPackageName(m_packageName);
        if (!m_applicationViewModes.isEmpty())
            desktopFileWriter.setApplicationViewMode(m_applicationViewModes.first());
        if (!m_homescreenViewModes.isEmpty())
            desktopFileWriter.setHomescreenViewMode(m_homescreenViewModes.first());

        if (!desktopFileWriter.write(m_widget))
            return false;

        m_nativeId = desktopFileWriter.desktopFilePath();
    }

    return true;
}

bool PackageUtils::writeInstallationScripts()
{
    // preinst script
#if defined(Q_OS_MAEMO5)
    // TODO: find a similar way to pop up license on other platform
    if (!m_license.isEmpty()) {
        m_preinstScript.append("#!/bin/sh\n"
                               "cat > /tmp/" + m_packageName + " <<EOF\n"
                               + m_license + "\n"
                               "EOF\n"
                               "maemo-confirm-text /tmp/" + m_packageName + "\n"
                               "retval=$?\n"
                               "rm /tmp/" + m_packageName + "\n"
                               "exit $retval");
    }
#endif

    // postinst script
    m_postinstScript.append("#!/bin/sh\n\n");
#if defined(Q_OS_MAEMO5)
    m_postinstScript.append("unzip '" + m_installationPath + QDir::separator() + QFileInfo(m_widget->widgetBundlePath()).fileName() + "' -d " + m_properties->installPath() + "\n");
#endif

#if !defined(Q_OS_MAEMO6)
    if (!m_iconSymbolicLinkPath.isEmpty()) {
        m_postinstScript.append("mkdir -p " + ICON_SYMBOLIC_LINK_DIR + "\n"
                                "if [ ! -f " + m_iconSymbolicLinkPath + " ]; then\n"
                                "ln -s " + m_iconPath + " " + m_iconSymbolicLinkPath + "\n"
                                "fi\n");
    }
#endif

    QString certificateAki = m_properties->certificateAki();
    if (certificateAki.isEmpty())
        certificateAki = "NONE";

    m_postinstScript.append("exec /usr/bin/webappregisterer " + m_properties->installPath() + " " + m_properties->id() + " " + m_nativeId + " " + certificateAki + " register\n"
                            "exit 0\n");

    return true;
}

bool PackageUtils::writeRemovalScripts()
{
    // postrm script
    m_postrmScript.append("#!/bin/sh\n\n");

#ifdef Q_OS_MAEMO5
    // remove unzipped files
    m_postrmScript.append("rm -rf /usr/share/wrt/data/webapps/" + m_properties->id() + "\n");
#endif

#if !defined(Q_OS_MAEMO6)
    if (!m_iconSymbolicLinkPath.isEmpty())
        m_postrmScript.append("rm -f " + m_iconSymbolicLinkPath + "\n");
#endif

    m_postrmScript.append("exec /usr/bin/webappregisterer " + m_properties->id() + " unregister\n\n"
                          "exit 0\n");

    return true;
}
