/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
 
#include "widgetinstaller.h"
#include "widgetinstaller_p.h"
#include "debianutils.h"
#include "featuremapping.h"
#include "storage.h"
#include "w3csettingskeys.h"
#include "wbenchmark.h"

#if defined(Q_OS_MAEMO5)
#include <dbus/dbus.h>
#include <hildon-mime.h>
#endif

#include <QDateTime>
#include <QProcess>
#include <QCoreApplication>
#include <sys/stat.h>
#include <sys/vfs.h>

WidgetInstallerPrivate::WidgetInstallerPrivate(SuperWidget* widget)
    : m_widget(widget), m_packageUtils(NULL), m_noInstall(false)
{
    // do nothing
}

WidgetInstallerPrivate::~WidgetInstallerPrivate()
{
    delete m_packageUtils;
    m_packageUtils = NULL;
}

WidgetInstallerPrivate::InstallationStatus WidgetInstallerPrivate::install(QString source, QString target, QString appId)
{
    WBM("Maemo specific installation");

    m_appId = appId;
    m_packageName = m_appId + "-wrt-widget";
    
    QStringList arguments = QCoreApplication::arguments();
    for (int i = 0; i < arguments.size(); ++i) {
        if (arguments.at(i) == "--path") {
            if(i+1 < arguments.size())
                m_outputPath = arguments.at(i+1);
        }
    }

    m_installationPath = target;
    if (!m_installationPath.endsWith(QDir::separator()))
        m_installationPath.append(QDir::separator());

#if defined(Q_OS_MAEMO5) || defined(Q_OS_MAEMO6)
    m_packageUtils = new DebianUtils(m_widget, m_packageName, source, m_installationPath, m_appId);
#else
    // TODO: fix for other platforms
#endif

    if (!m_packageUtils->createPackage()) {
        qCritical() << "Couldn't create package";
        emit m_widget->installationError(WidgetPlatformSpecificInstallFailed);
        return PackageCreationFailed;
    }

    qDebug() << "debcreated: " << m_packageUtils->packageFilePath();
    qDebug() << "installpath: " << m_installationPath;

    if (m_outputPath.isEmpty() && m_noInstall)
        m_outputPath = QDir::currentPath();

    if (!m_outputPath.isEmpty()) {
        if (!hasEnoughSpace(m_packageUtils->packageFilePath(), m_outputPath)) {
            qCritical() << "Not enough space";
            emit m_widget->installationError(WidgetInsufficientDiskSpace);
            return NotEnoughSpace;
        }

        m_outputPath.append(QDir::separator() + m_packageUtils->packageFileName());
        if (QFile::exists(m_outputPath))
            QFile::remove(m_outputPath);

        if (!QFile::rename(m_packageUtils->packageFilePath(), m_outputPath)) {
            qCritical() << "Couldn't move the package";
            emit m_widget->installationError(WidgetPlatformSpecificInstallFailed);
            return PackageMovingFailed;
        }

        return Success;
    }
    
    
    if (!installPackage()) {
        qCritical() << "Couldn't install Debian package";
        return PackageInstallationFailed;
    }
    
    WBM_BEGIN("secure storing widget files", secureStoring);
    if (!secureStoreWidget()) {
        qCritical() << "Couldn't securely store widget";
        emit m_widget->installationError(WidgetPlatformSpecificInstallFailed);
        return SecureStorageSetupFailed;
    }
    WBM_END(secureStoring);
    return Success;
}

bool WidgetInstallerPrivate::uninstall(QString appId)
{
    Q_UNUSED(appId)
    qWarning("WidgetInstallerPrivate::uninstall not removing debian package");
    return true;
}

void WidgetInstallerPrivate::setInstallationAttributes(WRT::InstallationAttributes attributes)
{
    if (attributes & WRT::NoInstall)
        m_noInstall = true;
}

#if defined(Q_OS_MAEMO5)
static gboolean quit_mainloop(gpointer data)
{
    g_main_loop_quit((GMainLoop*)data);
    return true;
}
#endif


bool WidgetInstallerPrivate::installPackage()
{
#if defined(Q_OS_MAEMO5)
    // TODO should use QDesktopServices::openUrl to open the Debian package
    qDebug() << "Install package with Application Manager launched by Hildon framework";

    DBusConnection* dbusConnection;
    DBusError dbusError;
    GMainLoop* mainloop = g_main_loop_new(NULL, FALSE);

    if (mainloop == NULL) {
      qCritical() << "Couldn't create mainloop";
      return false;
    }

    dbus_error_init(&dbusError);
    dbusConnection = dbus_bus_get(DBUS_BUS_SESSION, &dbusError);

    if (dbusConnection == NULL) {
        qCritical() << "Couldn't get DBus connection: " << dbusError.message;
        dbus_error_free(&dbusError);
        return false;
    }

    if (hildon_mime_open_file(dbusConnection, m_packageUtils->packageFilePath().toAscii().constData()) != 1) {
        qCritical() << "Cannot launch Application Manager";
        return false;
    }

    g_timeout_add(1000, quit_mainloop, (void*)mainloop);
    g_main_loop_run(mainloop);
#endif

    return true;
}

bool WidgetInstallerPrivate::secureStoreWidget()
{
    WRT::Storage* secureStorage = WRT::Storage::createInstance(m_packageName);
    if (secureStorage == NULL)
        return false;

    QString secureStorageFile = QDir::homePath() + QDir::separator() + ".maemosec-secure/s/" + m_packageName;

    // TODO: check on device if it allows to remove secure stogare file
    if (QFile::exists(secureStorageFile))
        QFile::remove(secureStorageFile);

    // add all files of the widget to sec storage
    QDir dir(m_installationPath);
    QStringList entries = dir.entryList(QDir::Dirs | QDir::NoDotAndDotDot);
    foreach (QString entry, entries) {
        QFile file(m_installationPath + entry);
        secureStorage->add(file);
    }
    return true;
}

bool WidgetInstallerPrivate::hasEnoughSpace(const QString& file, const QString& path) const
{
    unsigned long fileSize = QFile(file).size();

    unsigned long long availableSize = 0;
    struct statfs stats;
    statfs(path.toLocal8Bit(), &stats);
    availableSize = ((unsigned long long)stats.f_bavail) * ((unsigned long long)stats.f_bsize);

    if (availableSize < fileSize)
        return false;

    return true;
}
