/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#ifndef WGT_INSTALLER_P_MAEMO_H
#define WGT_INSTALLER_P_MAEMO_H


#include "SuperWidget.h"

#include <QString>
#include <QMap>

class PackageUtils;

/*!
  \class WidgetInstallerPrivate
  \brief Platform specific installations.

  It's only a QObject for Maemo6 as it requires to listen to the signals from Package Managers.
  */
class WidgetInstallerPrivate
{
    Q_DISABLE_COPY(WidgetInstallerPrivate)

public:
    /*!
      \enum WidgetInstallerPrivate::InstallationStatus
      \value Success The installation is successful. Note that as we can't track the return value of Application Manager on Maemo 5, it may still return this
      even if the installation actually fails.
      \value PackageCreationFailed Can't create the package.
      \value PackageInstallationFailed Can't install the generated package.
      \value PackageMovingFailed Can't move the generated package to the directory specified by the user.
      \value SecureStorageSetupFailed Can't put the installed widgets in secure storage.
      \value NotEnoughSpace Don't have enough space for the package.
     */
    enum InstallationStatus {
        Success = WidgetInstallSuccess,
        PackageCreationFailed,
        PackageInstallationFailed,
        PackageMovingFailed,
        SecureStorageSetupFailed,
        NotEnoughSpace
    };

    /*!
     Construct an object to install the \a widget.
     */
    explicit WidgetInstallerPrivate(SuperWidget* widget);

    /*!
     Destroy the object.
     */
    ~WidgetInstallerPrivate();

    /*!
     Install the unzipped widgets at \a source to \a target. The Unique ID is given as \a appId.
     */
    InstallationStatus install(QString source, QString target, QString appId);

    /*!
     Uninstall the widget with unique ID of \a appId.
     */
    static bool uninstall(QString appId);

    /*!
     Set the \a attributes for installation.
     */
    void setInstallationAttributes(WRT::InstallationAttributes attributes);

private:
    bool installPackage();
    bool secureStoreWidget();
    bool hasEnoughSpace(const QString& file, const QString& path) const;

    SuperWidget* m_widget;
    PackageUtils* m_packageUtils;
    QString m_appId;
    QString m_packageName;
    QString m_installationPath;
    bool m_noInstall;

    QString m_outputPath;
    QString m_productId;
};

#endif // WGT_INSTALLER_P_MAEMO_H
