/*
 * Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * This file is part of Qt Web Runtime.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "webapplocalizer.h"
#include "wrtsettings.h"
#include "WidgetUtilsLogs.h"
#include <QDir>
#include <QLocale>
#if defined(Q_OS_MAEMO6) || defined(Q_OS_MEEGO)
#include <MGConfItem>
#include <MLocale>
#endif

#if defined(USE_NEW_LOCALIZATION)
QHash<QString, QString> WebAppLocalizer::s_widgetExistsCache;
#else
QHash<QString, bool> WebAppLocalizer::s_widgetExistsCache;
#endif


WebAppLocalizer::WebAppLocalizer()
{

}

WebAppLocalizer::~WebAppLocalizer()
{
    s_widgetExistsCache.clear();
}

// function to retrieve localized version of a file
// parameters:
//    file            path to file that needs a localized version
//    absolute        if true provide absolute path, if false relative path
//    root            path to root of widget
// return:
//    QString         path of localized file
//
QString WebAppLocalizer::getLocalizedFile(const QString& file, const bool absolute, const QString& root)
{
    QString userAgentLang = platformLanguage();
#if defined(USE_NEW_LOCALIZATION)
    if (userAgentLang.isEmpty() )
        return file;

    QChar sep = QDir::separator();
    QString widgetRoot = QDir::fromNativeSeparators(root) + sep;

    // Find out relative path for with file name, without locales and language
    QString relativeFile = file;
    if (absolute) {
        relativeFile = QDir::fromNativeSeparators(file.mid(widgetRoot.length()));
        if (relativeFile.startsWith("locales/"))
            relativeFile = relativeFile.mid(relativeFile.indexOf("/", 8)+1);
        else
            relativeFile = file.mid(root.length()+1);
    }

    // Cache locales directory if not already done
    QString cacheDir = widgetRoot+"locales/";
    if (!s_widgetExistsCache.contains(cacheDir)) {
        QStringList dirList = QDir(cacheDir).entryList(QDir::CaseSensitive|QDir::Dirs|QDir::NoDotAndDotDot);
        foreach(QString dir, dirList) {
            // Have to cache real path because of case sensitivity
            s_widgetExistsCache.insert(cacheDir.toLower()+dir.toLower()+sep, cacheDir+dir+sep);
        }
        // just mark locales directory as cached and add root dir
        s_widgetExistsCache.insert(cacheDir, "cached");
        s_widgetExistsCache.insert(widgetRoot.toLower(), widgetRoot);
    }

    // Create pathlist where to search fallback files
    QStringList pathList(widgetRoot+"locales/"+userAgentLang+sep);
    QString lang(userAgentLang);
    while (lang.contains('-')) {
        lang.truncate(lang.lastIndexOf('-'));
         // add truncated localreglang to search path
        pathList << widgetRoot+"locales/"+lang+sep;
    }
    pathList << widgetRoot; // and widget root directory as a last search path

    // Return orgigal file as default result if localized not found
    QString result = file;

    // Search files
    foreach(QString searchPath, pathList) {
        QString cacheKey = searchPath.toLower();
        // Check cache does localized dir exist
        if (s_widgetExistsCache.contains(cacheKey)) {
            // Does the file exist?
            QString searchFile = s_widgetExistsCache.value(cacheKey) + relativeFile;
            if (QFile::exists(searchFile)) {
                if (absolute)
                    result = searchFile;
                else
                    result = searchFile.mid(widgetRoot.length());
                break; // We find it, let's go out from foreach
            }
        }
    }

    return QDir::toNativeSeparators(result);
#else
    // if file has userAgentLang, we are done
    //

    if (userAgentLang.isEmpty() )
        return file;



    // Default, nothing found
    QString result("");

    // List of languages
    QStringList langParts = userAgentLang.split("-");
    int langCount(langParts.count());

    QChar sep = QDir::separator();
    QString locales = "locales";
    QString nFile = QDir::toNativeSeparators(file);

    if (file.contains(QString("locales/" + userAgentLang),
          Qt::CaseInsensitive))
    {
        QFileInfo fInfo(nFile);
        nFile = fInfo.fileName();
    }

    // Construct a root path
    QString nRoot = QDir::toNativeSeparators(root);
    if (!nRoot.endsWith(sep)) {
        nRoot += sep;
    }

    // Take apart the path, handle absolute and relative paths
    //
    QString relPath("");
    if ( absolute ) {
        // Convert to relative path
        int len = nFile.length() - nRoot.length();
        if ( len >= 0 ) {
            relPath = nFile.right(len);
        }
        else {
            relPath = nFile;
        }
    }
    else {
        relPath = nFile;
    }

    // Split up the relative path, can be used for smarter algorithm
    //
    QStringList relPathParts = relPath.split(sep);
    int relCount = relPathParts.count() - 1; // last item is the file name
    QString fileName("");
    if ( relCount >=0 ) {
        fileName = relPathParts[relCount];
        relPathParts.removeLast();
    }

    // Lets try to open from best fit to worse fit, it we can open, we've found our answer!
    // Try to open locales folder + relative folder + file name if fail, move on
    //

    // Construct the relative path
    QString path(nRoot);
    path += locales;
    path += sep;

    // Check cache first for faster file access
    bool dirExists(false);
    QHash<QString, bool>::const_iterator it = s_widgetExistsCache.find(path);
    if (it != s_widgetExistsCache.end()) {
        dirExists = (bool) it.value();
    }
    else {
        // "locale" folder is case sensitive,
        // but the actual locale subfolder is not
        QDir dir(nRoot);
        QStringList localesList;
        localesList.append(locales);
        QStringList dirList = dir.entryList(localesList, QDir::CaseSensitive | QDir::Dirs);
        dirExists = (dirList.count() > 0);
        s_widgetExistsCache.insert(path, dirExists);
    }

    // Locales folder exists, so we can do the more expensive file checks
    //
    if (dirExists) {

        // Look for a locales + lang folder
        for (int i = langCount - 1; i >= 0; --i) {

            // build up the proposed localized path
            QString localized(path);

            // Chop extra '-', add separator
            for (int j = 0; j <= i; ++j) {
                localized += langParts[j];
                localized += "-";
            }

            localized.chop(1);

#ifndef Q_OS_SYMBIAN
            QDir langDir(path);
            QStringList langEntries = langDir.entryList(QDir::Dirs|QDir::NoDotAndDotDot);
            foreach(QString locName, langEntries) {
                QString fullLocName = path+locName;
                if (!fullLocName.compare(localized,Qt::CaseInsensitive)) {
                    localized = fullLocName;
                    break;
                }
            }
#endif
            localized += sep;

            // Add relative path, add file
            for (int l = 0; l < relCount; ++l) {
                localized += relPathParts[l];
                localized += sep;
            }

            // Check if we have looked for the folder before, can be cached
            // This will speed up cases where there are many files in the same folder
            //
            QHash<QString, bool>::const_iterator it = s_widgetExistsCache.find(localized);
            if (it != s_widgetExistsCache.end()) {
                dirExists = (bool) it.value();
            }
            else {
                dirExists = QFile::exists(localized);
                s_widgetExistsCache.insert(localized, dirExists);
            }

            // Up to the folder exists, check if file exists
            if (dirExists) {
                localized += fileName;

                // Lets look for the file. Optional file level cache
                if ( QFile::exists(localized) ) {
                    // We found something
                    if ( absolute ) {
                        result = localized;
                    }
                    else {
                        // relative path, just in case length is negative
                        int len(localized.length()-nRoot.length());
                        if ( len >=0 ) {
                            result = localized.right(len);
                        }
                        else {
                            result = localized;
                        }
                    }
                    break;
                }
            }
        } // for lang
    } // dir exists

    // By default, return the original file
    if ( result.length() == 0 ) {
        result = file;
    }
    return result;
#endif
}

QString WebAppLocalizer::findStartFile(const QString& contentSource, const QString& webAppRootDirectory)
{
    QStringList fileList;
    if (!contentSource.isEmpty()) {
        fileList.append(contentSource);
    }

    // Append default start files to list.
    fileList.append("index.htm");
    fileList.append("index.html");
    fileList.append("index.svg");
    fileList.append("index.xhtml");
    fileList.append("index.xht");

    foreach (QString fileName, fileList) {
        QString locFile = WebAppLocalizer::getLocalizedFile(fileName, false, webAppRootDirectory);
        LOG("Localized start file: " << locFile);
        if (!locFile.isEmpty() && QFile::exists(webAppRootDirectory + QDir::separator() + locFile)) {
            LOG("Localized start file: " << locFile);
            if (locFile.startsWith("locales/")) {
                // File name from locales/<lang>/ folder.
                int slashAfterLanguage = locFile.indexOf("/", 8);
                return locFile.mid(slashAfterLanguage + 1);
            } else {
                return locFile;
            }
        }
    }
    return QString();
}

QString WebAppLocalizer::platformLanguage()
{
    WRT::WrtSettings* settings = WRT::WrtSettings::createWrtSettings();
    QString userAgentLang = settings->valueAsString("UserAgentLanguage");
    if (userAgentLang.isEmpty() || userAgentLang.compare(" ") == 0) {
#if defined(Q_OS_MAEMO6) || defined(Q_OS_MEEGO) || defined(Q_OS_MAEMO5)

#if defined(Q_OS_MAEMO6) || defined(Q_OS_MEEGO)
        //Logic is similar to 'Language' control panel applet
        //Meego language is set using GConf
        const QString SettingsLanguage("/meegotouch/i18n/language");
        MGConfItem item (SettingsLanguage);
        QString resultLang = item.value().toString();
        if (resultLang.isEmpty()) {
            resultLang = MLocale().name(); // returning the default
        }
#else
        //Maemo5 language is set using 'setenv'
        QString resultLang = QString(qgetenv("LANG"));
        if (resultLang.isEmpty()) {
            resultLang = QLocale().name();
        }
#endif //#if defined(Q_OS_MAEMO6) || defined(Q_OS_MEEGO)

        // en means GB and not US, See bug: NB#167615
        if (resultLang == "en") {
            userAgentLang = "en_GB";
        } else {
            userAgentLang = QLocale(resultLang).name();
        }
        QLocale::setDefault(userAgentLang);
#endif
        userAgentLang = QLocale().name().toLower().replace(QString("_"),QString("-"));
    }
    return userAgentLang;
}

